import xml.etree.ElementTree as ET

import os
import os.path

from .baseGrant import Grant
from ..mkExceptions import BadGrant

class NSFGrant(Grant):
    def __init__(self, grantdDict, sFile):
        bad = False
        error = None
        originalName = os.path.basename(sFile)
        if grantdDict.get('AwardID', '') == '':
            bad = True
            error = BadGrant("Missing 'AwardID'")
            idValue = "NSF:{}".format(originalName)
        else:
            idValue = "NSF:{}".format(grantdDict.get('AwardID'))
        Grant.__init__(self, originalName, grantdDict, idValue, bad, error, sFile = sFile, sLine = 1)

    def getInvestigators(self, tags = None, seperator = ";", _getTag = False):
        """Returns a list of the names of investigators. The optional arguments are ignored.

        # Returns

        `list [str]`

        > A list of all the found investigator's names
        """
        if tags is None:
            tags = ['Investigator']
        elif isinstance(tags, str):
            tags = ['Investigator', tags]
        else:
            tags.append('Investigator')
        return super().getInvestigators(tags = tags, seperator = seperator, _getTag = _getTag)

    def getInstitutions(self, tags = None, seperator = ";", _getTag = False):
        """Returns a list with the names of the institution. The optional arguments are ignored

        # Returns

        `list [str]`

        > A list with 1 entry the name of the institution
        """
        if tags is None:
            tags = ['Institution']
        elif isinstance(tags, str):
            tags = ['Institution', tags]
        else:
            tags.append('Institution')
        return super().getInvestigators(tags = tags, seperator = seperator, _getTag = _getTag)

def isNSFfile(fileName, useFileName = True):
    if useFileName and not os.path.basename(fileName).endswith('.xml'):
        return False
    try:
        tree = ET.parse(fileName)
        root = tree.getroot()
        if len(root.findall('Award')) != 1:
            return False
        else:
            return True
    except (ET.ParseError, UnicodeError):
        return False

def parserNSFfile(fileName):
    error = None
    grantSet = set()
    grantDict = {}
    try:
        tree = ET.parse(fileName)
        top = tree.getroot().find('Award')
        #Organization is the only tag that goes 3 tags deep
        for org in top.findall('Organization'):
            directorate = org.find('Directorate')
            division = org.find('Division')
            if directorate is not None:
                directorateVal = []
                for subElement in directorate:
                    if subElement.text is not None:
                        directorateVal.append(subElement.text)
                if len(directorateVal) > 0:
                    if 'Directorate' not in grantDict:
                        grantDict['Directorate'] = []
                    grantDict['Directorate'].append('; '.join(directorateVal))
            if division is not None:
                divisionVal = []
                for subElement in division:
                    if subElement.text is not None:
                        divisionVal.append(subElement.text)
                if len(divisionVal) > 0:
                    if 'Division' not in grantDict:
                        grantDict['Division'] = []
                    grantDict['Division'].append('; '.join(divisionVal))
        for invest in top.findall('Investigator'):
            name = []
            other = []
            for subElement in invest:
                if subElement.text is None:
                    pass
                elif subElement.tag == 'FirstName':
                    name.insert(0, subElement.text)
                elif subElement.tag == 'LastName':
                    name.append(subElement.text)
                else:
                    other.append(subElement.text)
            investString = "{}; {}".format(' '.join(name), ', '.join(other))
            if 'Investigator' in grantDict:
                grantDict['Investigator'].append(investString)
            else:
                grantDict['Investigator'] = [investString]
        for xmlElement in top:
            if xmlElement.tag == 'Organization' or xmlElement.tag == 'Investigator':
                continue
            elif len(xmlElement) < 1:
                grantDict[xmlElement.tag] = xmlElement.text
            else:
                if xmlElement.tag not in grantDict:
                    grantDict[xmlElement.tag] = []
                elmComps = []
                for subElement in xmlElement:
                    if subElement.text is not None:
                        elmComps.append(subElement.text)
                grantDict[xmlElement.tag].append('; '.join(elmComps))
        grantSet.add(NSFGrant(grantDict, fileName))
    except (ET.ParseError, UnicodeError):
        if error is None:
            error = BadGrant("The file '{}' is having decoding issues. It may have been modifed since it was downloaded or not be a NSF grant file.".format(fileName))
    return grantSet, error



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
import unittest
import os.path
import os

import metaknowledge
import metaknowledge.medline


class TestMedline(unittest.TestCase):
    def setUp(self):
        metaknowledge.VERBOSE_MODE = False
        self.RC = metaknowledge.RecordCollection("metaknowledge/tests/medline_test.medline")
        self.R = self.RC.peek()

    def test_creation(self):
        Rstart = self.RC.peek()
        R = metaknowledge.MedlineRecord(Rstart._fieldDict)
        self.assertEqual(R, Rstart)
        with open("metaknowledge/tests/medline_test.medline") as f:
            f.readline()
            R = metaknowledge.MedlineRecord(f)
            self.assertEqual(R.id, 'PMID:26524502')
            s = f.read()
            R = metaknowledge.MedlineRecord(s)
            self.assertEqual(R.id, 'PMID:25802386')
        with self.assertRaises(TypeError):
            R = metaknowledge.MedlineRecord(12345678)
        R = metaknowledge.MedlineRecord("PMID- 25802386\njhgjhghjbgjhgjghhjgjh\nhdghjdfgjdfsgjh\n")
        self.assertTrue(R.bad)
        with self.assertRaises(metaknowledge.BadPubmedRecord):
            R.writeRecord('not a file')

    def test_isCollection(self):
        self.assertIsInstance(self.RC, metaknowledge.RecordCollection)

    def test_ismedline(self):
        self.assertIsInstance(self.R, metaknowledge.MedlineRecord)

    def test_bibWrite(self):
        fileName = "tempFile.bib.tmp"
        self.RC.writeBib(fileName)
        self.assertEqual(os.path.getsize(fileName), 606182)
        self.RC.writeBib(fileName, wosMode = True, reducedOutput = True)
        self.assertEqual(os.path.getsize(fileName), 456151)
        os.remove("tempFile.bib.tmp")

    def test_specials(self):
        for R in self.RC:
            for s in metaknowledge.medline.medlineSpecialTagToFunc.keys():
                self.assertIsInstance(R.get(s), (str, type(None), list, int, metaknowledge.Citation))

    def test_allFields(self):
        for R in self.RC:
            for k,v in R.items():
                self.assertIsInstance(k, str)
                self.assertIsInstance(v, (str, list, dict))

    def test_write(self):
        fileName = 'tempFile.medline.tmp'
        self.RC.writeFile(fileName)
        self.assertEqual(os.path.getsize(fileName), os.path.getsize("metaknowledge/tests/medline_test.medline") + 526) #Not quite identical
        os.remove(fileName)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
import unittest
import unittest.mock
import builtins
import importlib
import sys
import metaknowledge.constants

class TestConstants(unittest.TestCase):
    def test_VerboseMode(self):
        self.assertFalse(metaknowledge.constants.isInteractive())
        sys.stdout.isatty = lambda : True
        self.assertTrue(metaknowledge.constants.isInteractive())
        class ImportMock(unittest.mock.Mock):
            def __call__(self, *args, **kwargs):
                if args[0] == 'threading':
                    raise ImportError
                else:
                    return importlib.__import__(*args, **kwargs)
        with unittest.mock.patch('builtins.__import__', new_callable = ImportMock):#, NoThreadingImport):
        #builtins.__import__ =
            self.assertFalse(metaknowledge.constants.isInteractive()) #This will fail for setup.py test
            #Failure for setup.py is what is supposed to happen as that would be an interactive enviroment



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import abc
try:
    import collections.abc
except ImportError:
    import collections
    collections.abc = collections
import re

from .mkExceptions import BadCitation
from .journalAbbreviations import getj9dict, abrevDBname, manualDBname, addToDB

import metaknowledge

#For journalAbbreviations, to reduce the number of times we read the dict
abbrevDict = None

class Citation(collections.abc.Hashable):
    """A class to hold citation strings and allow for comparison between them.

    The initializer takes in a string representing a WOS citation in the form:

        Author, Year, Journal, Volume, Page, DOI

    `Author` is the author's name in the form of first last name first initial sometimes followed by a period.
    `Year` is the year of publication.
    `Journal` being the 29-Character Source Abbreviation of the journal.
    `Volume` is the volume number(s) of the publication preceded by a V
    `Page` is the page number the record starts on
    `DOI` is the DOI number of the cited record preceeded by the letters `'DOI'`
    Combined they look like:

        Nunez R., 1998, MATH COGNITION, V4, P85, DOI 10.1080/135467998387343

    **Note**: any of the fields have been known to be missing and the requirements for the fields are not always met. If something is in the source string that cannot be interpreted as any of these it is put in the `misc` attribute. That is the reason to use this class, it gracefully handles missing information while still allowing for  comparison between WOS citation strings.

    # Customizations

    Citation's hashing and equality checking are based on [ID()](#metaknowledge.citation.Citation.ID) and use the values of `author`, `year` and `journal`.

    When converted to a string a Citation will return the original string.

    # Attributes

    As noted above, citations are considered to be divided into six distinct fields (`Author`, `Year`, `Journal`, `Volume`, `Page` and `DOI`) with a seventh `misc` for anything not in those. Records thus have an attribute with a name corresponding to each `author`, `year`, `journal`, `V`, `P`, `DOI` and `misc` respectively. These are created if there is anything in the field. So a `Citation` created from the string: `'Nunez R., 1998, MATH COGNITION'` would have `author`, `year` and `journal` defined. While one from `'Nunez R.'` would have only the attribute `misc`.

    If the parsing of a citation string fails the attribute `bad` is set to `True` and the attribute `error` is created to contain said error, which is a [BadCitation](../exceptions/index.html#metaknowledge.mkExceptions.BadCitation) object. If no errors occur `bad` is `False`.

    The attribute `original` is the unmodified string (_cite_) given to create the Citation, it can also be accessed by converting to a string, e.g. with `str()`.

    # \_\_Init\_\_

    Citations can be created by [Records](./Record.html#metaknowledge.Record) or by giving the initializer a string containing a WOS style citation.

    # Parameters

    _cite_ : `str`

    > A str containing a WOS style citation.
    """
    wosCiteRegex = re.compile(r"([^0-9,][^,]+)?(, )?(-?[0-9]{1,5})?(, )?([^,]+)?(, (V[^,]+))?(, (P[^,]+))?($|, DOI (.+)|((.+?)(, DOI (.+))?))")

    scopusCiteRegex = re.compile(r"([\w\- ]+,? [\w\.\-]+\.), (([\w\- ]+,? [\w\-\.]+\., )+)?([^(]+? )?\((\d{2,5})\) ([^,]+)(, ([\w-]+)( \(([\w-]*)\))?)?(, P?P\. ([\w-]*))?(.*)")


    #(.*?, )?[^,(]*\((\d{2,5}\))[^,]*(, [^,(]*( \(\d+\))?)?(, (p?p\. .*))?")

    def __init__(self, cite, scopusMode = False):
        #save original
        #setup attributes
        #Nunez R., 1998, MATH COGNITION, V4, P85, DOI 10.1080/135467998387343
        #Author, Year, Journal, Volume, Page, DOI
        if scopusMode:
            regex = re.match(self.scopusCiteRegex, cite.upper())
            if regex is None:
                self.bad = True
                self.error = BadCitation("Regex parsing failed on a Scopus Citation this means the Citation is likely for a non-journal.")
                self._id = cite
            else:
                self.author = regex.group(1)
                self.extraAuthors = regex.group(2)
                self.name = regex.group(4)
                self.year = int(regex.group(5))
                self.journal = regex.group(6)
                self.V = regex.group(8)
                self.issue = regex.group(10)
                self.P = regex.group(12)
                self.misc = regex.group(13)
                if self.author and self.journal and self.year:
                    self.bad = False
                    self.error = None
                    self._id =  "{0}, {1}, {2}".format(self.author, self.year, self.journal)
                else:
                    self.bad = True
                    self.error = BadCitation("Not a complete set of author, year and journal")
                    atrLst = []
                    if self.author:
                        atrLst.append(self.author)
                    if self.year:
                        atrLst.append(str(self.year))
                    if self.journal:
                        atrLst.append(self.journal)
                    self._id =  ', '.join(atrLst)
        else:
            regex = re.match(self.wosCiteRegex, cite.upper())
            if regex is None:
                self.bad = True
                self.error = BadCitation("Regex parsing failed.")
            try:
                self.author = regex.group(1).replace('.', '').title()
            except AttributeError:
                self.author = None
            try:
                self.year = int(regex.group(3))
            except TypeError:
                self.year = None
            self.journal = regex.group(5)
            self.V = regex.group(7)
            self.P = regex.group(9)
            self.DOI = regex.group(11)
            if regex.group(12) is not None:
                self.misc = regex.group(12)
                self.DOI = regex.group(15)
                self.bad = True
                self.error = BadCitation("The citation did not fully match the expected pattern")
                atrLst = []
                if self.author:
                    atrLst.append(self.author)
                if self.year:
                    atrLst.append(str(self.year))
                if self.journal:
                    atrLst.append(self.journal)
                self._id =  ', '.join(atrLst)
            elif self.author is None or self.year is None or self.journal is None:
                self.bad = True
                self.misc = None
                self.error = BadCitation("Not a complete set of author, year and journal")
                atrLst = []
                if self.author:
                    atrLst.append(self.author)
                if self.year:
                    atrLst.append(str(self.year))
                if self.journal:
                    atrLst.append(self.journal)
                self._id =  ', '.join(atrLst)
            else:
                self.bad = False
                self.error = None
                self.misc = None
                self._id =  "{0}, {1}, {2}".format(self.author, self.year, self.journal)
        if not metaknowledge.FAST_CITES:
            self.original = cite
        else:
            self = self._id

    def __str__(self):
        """
        returns the original string
        """
        if metaknowledge.FAST_CITES:
            return self.ID()
        else:
            return self.original

    def __repr__(self):
        """
        the representation of the Citation is its original form
        """
        if metaknowledge.FAST_CITES:
            return "<metaknowledge.{} object {}>".format(type(self).__name__, self.ID())
        else:
            return "<metaknowledge.{} object {}>".format(type(self).__name__, self.original)
    #@profile
    def __hash__(self):
        """
        A hash for Citation that should be equal to the hash of other citations that are equal to it. Based on the values returned by [ID()](#metaknowledge.citation.Citation.ID).
        """
        try:
            return self._hash
        except AttributeError:
            self._hash = hash(self.ID())
            return self._hash
    #@profile
    def __eq__(self, other):
        """
        First checks DOI for equality then checks each attribute if any are not equal False is returned
        """
        if not isinstance(other, Citation):
            return NotImplemented
        return hash(self) == hash(other)


    def isAnonymous(self):
        """
        Checks if the author is given as `'[ANONYMOUS]'` and returns `True` if so.

        # Returns

        `bool`

        > `True` if the author is `'[ANONYMOUS]'` otherwise `False`.
        """
        try:
            return self.author == "[Anonymous]"
        except AttributeError:
            return True

    def ID(self):
        """
        Returns all of `author`, `year` and `journal` available separated by `' ,'`. It is for shortening labels when creating networks as the resultant strings are often unique. [Extra()](#metaknowledge.citation.Citation.Extra) gets everything not returned by **ID**().

        This is also used for hashing and equality checking.

        # Returns

        `str`

        > A string to use as the ID of a node.
        """
        return self._id

    def allButDOI(self):
        """
        Returns a string of the normalized values from the Citation excluding the DOI number. Equivalent to getting the ID with [ID()](#metaknowledge.citation.Citation.ID) then appending the extra values from [Extra()](#metaknowledge.citation.Citation.Extra) and then removing the substring containing the DOI number.

        # Returns

        `str`

        > A string containing the data of the Citation.
        """
        extraTags = ['extraAuthors', 'V', 'issue', 'P', 'misc']
        s = self.ID()
        extras = []
        for tag in extraTags:
            if getattr(self, tag, False):
                extras.append(str(getattr(self, tag)))
        if len(extras) > 0:
            return "{0}, {1}".format(s, ', '.join(extras))
        else:
            return s

    def Extra(self):
        """
        Returns any `V`, `P`, `DOI` or `misc` values as a string. These are all the values not returned by [ID()](#metaknowledge.citation.Citation.ID), they are separated by `' ,'`.

        # Returns

        `str`

        > A string containing the data not in the ID of the `Citation`.
        """
        extraTags = ['V', 'P', 'DOI', 'misc']
        retVal = ""
        for tag in extraTags:
            if getattr(self, tag):
                retVal += getattr(self, tag) + ', '
        if len(retVal) > 2:
            return retVal[:-2]
        else:
            return retVal

    def isJournal(self, dbname = abrevDBname, manualDB = manualDBname, returnDict ='both', checkIfExcluded = False):
        """Returns `True` if the `Citation`'s `journal` field is a journal abbreviation from the WOS listing found at [http://images.webofknowledge.com/WOK46/help/WOS/A_abrvjt.html](http://images.webofknowledge.com/WOK46/help/WOS/A_abrvjt.html), i.e. checks if the citation is citing a journal.

        **Note**: Requires the [j9Abbreviations](../modules/journalAbbreviations.html#metaknowledge.journalAbbreviations.backend.getj9dict) database file and will raise an error if it cannot be found.

        **Note**: All parameters are used for getting the data base with [getj9dict](../modules/journalAbbreviations.html#metaknowledge.journalAbbreviations.backend.getj9dict).

        # Parameters

        _dbname_ : `optional [str]`

        > The name of the downloaded database file, the default is determined at run time. It is recommended that this remain untouched.

        _manualDB_ : `optional [str]`

        > The name of the manually created database file, the default is determined at run time. It is recommended that this remain untouched.

        _returnDict_ : `optional [str]`

        > default `'both'`, can be used to get both databases or only one  with `'WOS'` or `'manual'`.

        # Returns

        `bool`

        > `True` if the `Citation` is for a journal
        """
        global abbrevDict
        if abbrevDict is None:
            abbrevDict = getj9dict(dbname = dbname, manualDB = manualDB, returnDict = returnDict)
        if not hasattr(self, 'journal'):
            return False
        elif checkIfExcluded and self.journal:
            try:
                if abbrevDict.get(self.journal, [True])[0]:
                    return False
                else:
                    return True
            except IndexError:
                return False
        else:
            if self.journal:
                dictVal = abbrevDict.get(self.journal, [b''])[0]
                if dictVal:
                    return dictVal
                else:
                    return False
            else:
                return False

    def FullJournalName(self):
        """Returns the full name of the Citation's journal field. Requires the [j9Abbreviations](../modules/journalAbbreviations.html#metaknowledge.journalAbbreviations.backend.getj9dict) database file.

        **Note**: Requires the [j9Abbreviations](../modules/journalAbbreviations.html#metaknowledge.journalAbbreviations.backend.getj9dict) database file and will raise an error if it cannot be found.

        # Returns

        `str`

        > The first full name given for the journal of the Citation (or the first name in the WOS list if multiple names exist), if there is not one then `None` is returned
        """
        global abbrevDict
        if abbrevDict is None:
            abbrevDict = getj9dict()
        if self.isJournal():
            return abbrevDict[self.journal][0]
        else:
            return None

    def addToDB(self, manualName = None, manualDB = manualDBname, invert = False):
        """Adds the journal of this Citation to the user created database of journals. This will cause [isJournal()](#metaknowledge.citation.Citation.isJournal) to return `True` for this Citation and all others with its `journal`.

        **Note**: Requires the [j9Abbreviations](../modules/journalAbbreviations.html#metaknowledge.journalAbbreviations.backend.getj9dict) database file and will raise an error if it cannot be found.

        # Parameters

        _manualName_ : `optional [str]`

        > Default `None`, the full name of journal to use. If not provided the full name will be the same as the abbreviation.

        _manualDB_ : `optional [str]`

        > The name of the manually created database file, the default is determined at run time. It is recommended that this remain untouched.

        _invert_ : `optional [bool]`

        > Default `False`, if `True` the journal will be removed instead of added
        """
        try:
            if invert:
                d = {self.journal : ''}
            elif manualName is None:
                d = {self.journal : self.journal}
            else:
                d = {self.journal : manualName}
            addToDB(abbr = d, dbname = manualDB)
        except KeyError:
            raise KeyError("This citation does not have a journal field.")
        else:
            abbrevDict.update(d)

def filterNonJournals(citesLst, invert = False):
    """Removes the `Citations` from _citesLst_ that are not journals

    # Parameters

    _citesLst_ : `list [Citation]`

    > A list of citations to be filtered

    _invert_ : `optional [bool]`

    > Default `False`, if `True` non-journals will be kept instead of journals

    # Returns

    `list [Citation]`

    > A filtered list of Citations from _citesLst_
    """

    retCites = []
    for c in citesLst:
        if c.isJournal():
            if not invert:
                retCites.append(c)
        elif invert:
            retCites.append(c)
    return retCites



try:
    import collections.abc
except ImportError:
    import collections
    collections.abc = collections
import csv
import os

from ..mkRecord import Record
from ..mkExceptions import BadGrant



class Grant(Record, collections.abc.MutableMapping):

    #Overwriting Record's attribute
    _documented = []

    def __init__(self, original, grantdDict, idValue, bad, error, sFile = "", sLine = 0):
        self.original = original
        Record.__init__(self, grantdDict, idValue, bad, error, sFile = sFile, sLine = sLine)

    def __setitem__(self, key, value):
        if isinstance(key, str):
            self._fieldDict.__setitem__(key, value)
        else:
            raise KeyError("{} is not a valid key, all keys must be strings.".format(key))

    def __delitem__(self, key):
        self._fieldDict.__delitem__(key)

    def getInvestigators(self, tags = None, seperator = ";", _getTag = False):
        """Returns a list of the names of investigators. This is done by looking (in order) for any of fields in _tags_ and splitting the strings on _seperator_. If no strings are found an empty list will be returned.

        *Note* for some Grants `getInvestigators` has been overwritten and will ignore the arguments and simply provide the investigators.

        # Parameters

        _tags_ : `optional list[str]`

        > A list of the tags to look for investigators in

        _seperator_ : `optional str`

        > The string that separators each investigators name  within the column

        # Returns

        `list [str]`

        > A list of all the found investigator's names
        """
        #By default we don't know which field has the investigators
        investVal = []
        retTag = None
        if tags is not None:
            if not isinstance(tags, list):
                tags = [tags]
            for tag in tags:
                try:
                    tval = self[tag].split(seperator)
                    if _getTag:
                        investVal += [(t.strip(), tag) for t in tval]
                    else:
                        investVal += [t.strip() for t in tval]
                except KeyError:
                    pass
                except AttributeError:
                    tval = [auth.split(seperator)[0] for auth in self[tag]]
                    if _getTag:
                        investVal += [(t.strip(), tag) for t in tval]
                    else:
                        investVal += [t.strip() for t in tval]
        return investVal

    def getInstitutions(self, tags = None, seperator = ";", _getTag = False):
        """Returns a list of the names of institutions. This is done by looking (in order) for any of fields in _tags_ and splitting the strings on _seperator_ (in case of multiple institutions). If no strings are found an empty list will be returned.

        *Note* for some Grants `getInstitutions` has been overwritten and will ignore the arguments and simply provide the investigators.

        # Parameters

        _tags_ : `optional list[str]`

        > A list of the tags to look for institutions in

        _seperator_ : `optional str`

        > The string that separators each institutions name within the column

        # Returns

        `list [str]`

        > A list of all the found institution's names
        """
        return self.getInvestigators(tags = tags, seperator = seperator, _getTag = _getTag)

    def update(self, other):
        """Adds all the tag-entry pairs from _other_ to the `Grant`. If there is a conflict _other_ takes precedence.

        # Parameters

        _other_ : `Grant`

        > Another `Grant` of the same type as _self_
        """
        if type(self) != type(other):
            return NotImplemented
        else:
            if other.bad:
                self.error = other.error
                self.bad = True
            self._fieldDict.update(other._fieldDict)

def csvAndLinesReader(enumeratedFile, *csvArgs, **csvKwargs):
    currentData = {
    'currentLineNum' : -1,
    'currentLineString' : '',
    }
    def readerWithSideEffects(target, datDict):
        while True:
            datDict['currentLineNum'], datDict['currentLineString'] = next(target)
            yield datDict['currentLineString']
    reader = csv.DictReader(readerWithSideEffects(enumeratedFile, currentData), *csvArgs, **csvKwargs)
    readerIter = reader.__iter__()
    while True:
        row = next(readerIter)
        yield currentData['currentLineNum'], currentData['currentLineString'], row

class FallbackGrant(Grant):
    """A subclass of [Grant](./grants.html#metaknowledge.grants.Grant), it has the same attributes and is returned from the fall back constructor for grants.
    """
    #Making it a subclass so that Grant is never used raw
    #Also make interface simpler
    def __init__(self, original, grantdDict, sFile = "", sLine = 0):
        #We do not known anything about the structure of the grant so there is nothing to check about causing an error
        #The id needs to be unique so hashing the original will always give us that
        idValue = "{}-l:{}-{:0=20}".format(os.path.basename(sFile), sLine, hash(original))
        Grant.__init__(self, original, grantdDict, idValue, False, None, sFile = sFile, sLine = sLine)

def isFallbackGrantFile(fileName, useFileName = True, encoding = 'latin-1', dialect = 'excel'):
    if useFileName:
        if not fileName.endswith('csv'):
            return False
    try:
        #Try to open it
        with open(fileName, 'r', encoding = encoding) as openfile:
            #See if csv likes it
            reader = csv.DictReader(openfile, fieldnames = None, dialect = dialect)

            #Check that it is not a scopus file
            if len(set(reader.fieldnames) & {'Conference location', 'Conference code', 'ISSN', 'ISBN', 'CODEN', 'PubMed ID', 'Language of Original Document', 'Abbreviated Source Title', 'Document Type', 'Source', 'EID'}) == 11:
                return False
            #Check that every row can be read
            length = 0
            for row in reader:
                length += 1
                #Just wanted to put something here that analysis the row
                #There is probably a better check
                if set(row.keys()) != set(reader.fieldnames):
                    return False
            #Check that there are rows to read
            if length < 1:
                return False
    except (StopIteration, UnicodeDecodeError, csv.Error, TypeError):
        #If any of theses exceptions are raised the nit is defintly not as csv file
        #We do not want to catch everything though as there could be an issue with the code
        return False
    else:
        #IF nothing caused an issue return True
        return True

def parserFallbackGrantFile(fileName, encoding = 'latin-1', dialect = 'excel'):
    #Declare the returns out side of the block to show they are accessible everywhere inside it and so if there are issues with their creation it will no cause a problem with returning them
    grantSet = set()
    error = None
    try:
        with open(fileName, 'r', encoding = encoding) as openfile:
            f = enumerate(openfile, start = 1)
            reader = csvAndLinesReader(f, fieldnames = None, dialect = dialect)
            for lineNum, lineString, lineDict in reader:
                grantSet.add(FallbackGrant(lineString, lineDict, sFile = fileName, sLine = lineNum))
    except Exception:
        if error is None:
            error = BadGrant("The file '{}' is having decoding issues. It may have been modifed since it was downloaded or not be a CIHR grant file.".format(fileName))
    except KeyboardInterrupt as e:
        error = e
    finally:
        if isinstance(error, KeyboardInterrupt):
            raise error
        return grantSet, error



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
import unittest
import metaknowledge

import os

class TestProQuest(unittest.TestCase):

    def setUp(self):
        metaknowledge.VERBOSE_MODE = False
        self.RC = metaknowledge.RecordCollection("metaknowledge/tests/ProQuest_TestFile.testtxt")
        self.R = self.RC.peek()

    def test_isCollection(self):
        self.assertIsInstance(self.RC, metaknowledge.RecordCollection)

    def test_isProQuest(self):
        self.assertIsInstance(self.R, metaknowledge.ProQuestRecord)

    def test_specials(self):
        for R in self.RC:
            for s in metaknowledge.proquest.proQuestSpecialTagToFunc.keys():
                self.assertIsInstance(R.get(s), (str, type(None), list, int, metaknowledge.Citation))

    def test_allFields(self):
        for R in self.RC:
            for k,v in R.items():
                self.assertIsInstance(k, str)
                self.assertIsInstance(v, (str, list, int))

    def test_graphs(self):
        self.assertEqual(metaknowledge.graphStats(self.RC.networkMultiMode(self.RC.tags()), sentenceString = True), "The graph has 1928 nodes, 50833 edges, 0 isolates, 114 self loops, a density of 0.0273952 and a transitivity of 0.0815136")

    def test_write(self):
        #No writer currently implemented
        pass



try:
    import collections.abc
except ImportError:
    import collections
    collections.abc = collections

from .mkExceptions import UnknownFile

from .grants.cihrGrant import parserCIHRfile, isCIHRfile
from .grants.nsercGrant import parserNSERCfile, isNSERCfile
from .grants.nsfGrant import parserNSFfile, isNSFfile
from .grants.baseGrant import parserFallbackGrantFile, isFallbackGrantFile

from .WOS.wosHandlers import isWOSFile, wosParser
from .medline.medlineHandlers import isMedlineFile, medlineParser
from .proquest.proQuestHandlers import isProQuestFile, proQuestParser
from .scopus.scopusHandlers import isScopusFile, scopusParser

ProccessorTuple = collections.namedtuple("ProccessorTuple", ("type", "processor", "detector"))

def unrecognizedFileHandler(fileName):
    raise UnknownFile("'{}' is not recognized my metaknowledge.".format(fileName))

grantProcessors = [
    ProccessorTuple("NSFGrant", parserNSFfile, isNSFfile),
    ProccessorTuple("CIHRGrant", parserCIHRfile, isCIHRfile),
    ProccessorTuple("NSERCGrant", parserNSERCfile, isNSERCfile),
    ProccessorTuple("FallbackGrant", parserFallbackGrantFile, isFallbackGrantFile),
    #Raises exception if reached, to indicate the end of the list
    #This simplifes things at the other end
    ProccessorTuple("Invalid File", None, unrecognizedFileHandler),
]

recordHandlers = [
    ProccessorTuple("WOSRecord", wosParser, isWOSFile),
    ProccessorTuple("MedlineRecord", medlineParser, isMedlineFile),
    ProccessorTuple("ProQuestRecord", proQuestParser, isProQuestFile),
    ProccessorTuple("ScopusRecord", scopusParser, isScopusFile),
    #Raises exception if reached, to indicate the end of the list
    #This simplifes things at the other end
    ProccessorTuple("Invalid File", None, unrecognizedFileHandler),
]



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import unittest
import metaknowledge

class TestCitation(unittest.TestCase):
    def setUp(self):
        self.Cite = metaknowledge.Citation("John D., 2015, TOPICS IN COGNITIVE SCIENCE, V1, P1, DOI 0.1063/1.1695064")

    def test_citation_author(self):
        self.assertEqual(self.Cite.author, "John D")

    def test_citation_year(self):
        self.assertEqual(self.Cite.year, 2015)

    def test_citation_journal(self):
        self.assertEqual(self.Cite.journal, "TOPICS IN COGNITIVE SCIENCE")

    def test_citation_v(self):
        self.assertEqual(self.Cite.V, "V1")

    def test_citation_p(self):
        self.assertEqual(self.Cite.P, "P1")

    def test_citation_DOI(self):
        self.assertEqual(self.Cite.DOI, "0.1063/1.1695064")

    def test_citation_id(self):
        self.assertEqual(self.Cite.ID(), "John D, 2015, TOPICS IN COGNITIVE SCIENCE")

    def test_citation_str(self):
        self.assertEqual(str(self.Cite), "John D., 2015, TOPICS IN COGNITIVE SCIENCE, V1, P1, DOI 0.1063/1.1695064")

    def test_citation_extra(self):
        self.assertEqual(self.Cite.Extra(), "V1, P1, 0.1063/1.1695064")

    def test_citation_badDetection(self):
        self.assertTrue(metaknowledge.Citation("").bad)

    def test_citation_equality(self):
        c1 = metaknowledge.Citation("John D., 2015, TOPICS IN COGNITIVE SCIENCE, P1, DOI 0.1063/1.1695064")
        c2 = metaknowledge.Citation("John D., 2015, TOPICS IN COGNITIVE SCIENCE, V1, P1")
        c3 = metaknowledge.Citation("John D., 2015, TOPICS IN COGNITIVE SCIENCE, V1, P2")
        self.assertTrue(c1 == self.Cite)
        self.assertTrue(c2 == self.Cite)
        self.assertFalse(c1 != c2)
        self.assertFalse(c3 != c1)

    def test_citation_hash(self):
        self.assertTrue(bool(hash(self.Cite)))
        self.assertTrue(bool(hash(metaknowledge.Citation("John D., 2015, TOPICS IN COGNITIVE SCIENCE, V1, P1"))))
        self.assertTrue(bool(hash(metaknowledge.Citation("John D., 2015"))))

    def test_citation_badLength(self):
        c = metaknowledge.Citation("ab, c")
        self.assertTrue(c.bad)
        self.assertEqual(str(c.error), "Not a complete set of author, year and journal")
        self.assertEqual(c.Extra(),'')
        self.assertEqual(c.author,'Ab')
        self.assertEqual(c.ID(),'Ab, C')

    def test_citation_badNumbers(self):
        c = metaknowledge.Citation("1, 2, 3, 4")
        self.assertTrue(c.bad)
        self.assertEqual(c.ID(), '1, 2')
        self.assertEqual(str(c.error), "The citation did not fully match the expected pattern")



import os.path
import csv

from .baseGrant import Grant, csvAndLinesReader
from ..mkExceptions import BadGrant

class CIHRGrant(Grant):
    def __init__(self, original, grantdDict, sFile, sLine):
        bad = False
        error = None
        if grantdDict.get('PI Names', '') == '':
            bad = True
            error = BadGrant("Missing 'PI Names'")

        #Source file - line number - 6 character long numeric hash
        idValue = "{}-l:{}-{:0=20}".format(os.path.basename(sFile), sLine, hash(original))

        Grant.__init__(self, original, grantdDict, idValue, bad, error, sFile = sFile, sLine = sLine)

def isCIHRfile(fileName, useFileName = True):
    if useFileName and not os.path.basename(fileName).startswith('cihr_'):
        return False
    try:
        with open(fileName, 'r', encoding = 'latin-1') as openfile:
            if not openfile.readline().startswith('Search Criteria'):
                return False
            elif not openfile.readline().endswith(',,,,,,,,,\n'):
                return False
            elif not openfile.readline().endswith(',,,,,,,,,\n'):
                return False
            reader = csv.DictReader(openfile, fieldnames = None, dialect = 'excel')
            for row in reader:
                if 'PI Names' not in row:
                    return False
    except (StopIteration, UnicodeDecodeError):
        return False
    else:
        return True

def parserCIHRfile(fileName):
    grantSet = set()
    error = None
    try:
        with open(fileName, 'r', encoding = 'latin-1') as openfile:
            f = enumerate(openfile, start = 1)
            next(f)
            next(f)
            next(f)
            reader = csvAndLinesReader(f, fieldnames = None, dialect = 'excel')
            for lineNum, lineString, lineDict in reader:
                grantSet.add(CIHRGrant(lineString, lineDict, sFile = fileName, sLine = lineNum))
    except Exception:
        if error is None:
            error = BadGrant("The file '{}' is having decoding issues. It may have been modifed since it was downloaded or not be a CIHR grant file.".format(fileName))
    except KeyboardInterrupt as e:
        error = e
    finally:
        if isinstance(error, KeyboardInterrupt):
            raise error
        return grantSet, error



import itertools

from .recordWOS import WOSRecord
from ..mkExceptions import cacheError, BadWOSFile, BadWOSRecord

def isWOSFile(infile, checkedLines = 3):
    """Determines if _infile_ is the path to a WOS file. A file is considerd to be a WOS file if it has the correct encoding (`utf-8` with a BOM) and within the first _checkedLines_ a line starts with `"VR 1.0"`.

    # Parameters

    _infile_ : `str`

    > The path to the targets file

    _checkedLines_ : `optional [int]`

    > default 2, the number of lines to check for the header

    # Returns

    `bool`

    > `True` if the file is a WOS file
    """
    try:
        with open(infile, 'r', encoding='utf-8-sig') as openfile:
            f = enumerate(openfile, start = 0)
            for i in range(checkedLines):
                if "VR 1.0" in f.__next__()[1]:
                    return True
    except (StopIteration, UnicodeDecodeError):
        return False
    else:
        return False

def wosParser(isifile):
    """This is a function that is used to create [RecordCollections](../classes/RecordCollection.html#metaknowledge.RecordCollection) from files.

    **wosParser**() reads the file given by the path isifile, checks that the header is correct then reads until it reaches EF. All WOS records it encounters are parsed with [recordParser()](#metaknowledge.WOS.recordWOS.recordParser) and converted into [Records](../classes/Record.html#metaknowledge.Record). A list of these `Records` is returned.

    `BadWOSFile` is raised if an issue is found with the file.

    # Parameters

    _isifile_ : `str`

    > The path to the target file

    # Returns

    `List[Record]`

    > All the `Records` found in _isifile_
    """
    plst = set()
    error = None
    try:
        with open(isifile, 'r', encoding='utf-8-sig') as openfile:
            f = enumerate(openfile, start = 0)
            while "VR 1.0" not in f.__next__()[1]:
                pass
            notEnd = True
            while notEnd:
                line = f.__next__()
                if line[1] == '':
                    error =  BadWOSFile("'{}' does not have an 'EF', lines 1 to {} were checked".format(isifile, line[0] + 1))
                elif line[1].isspace():
                    continue
                elif 'EF' in line[1][:2]:
                    notEnd = False
                    continue
                else:
                    try:
                        plst.add(WOSRecord(itertools.chain([line], f), sFile = isifile, sLine = line[0]))
                    except BadWOSFile as e:
                        try:
                            s = f.__next__()[1]
                            while s[:2] != 'ER':
                                s = f.__next__()[1]
                        except:
                            error =  BadWOSFile("The file {} was not terminated corrrectly caused the following error:\n{}".format(isifile, str(e)))
            try:
                f.__next__()
            except StopIteration:
                pass
            else:
                error =  BadWOSFile("EF not at end of " + isifile)
    except UnicodeDecodeError:
        try:
            error =  BadWOSFile("'{}' has a unicode issue on line: {}.".format(isifile, f.__next__()[0]))
        except:
            #Fallback needed incase f.__next__() causes issues
            error =  BadWOSFile("'{}' has a unicode issue. Probably when being opened or possibly on the first line".format(isifile))
    except StopIteration:
        error =  BadWOSFile("The file '{}' ends before EF was found".format(isifile))
    except KeyboardInterrupt as e:
        error = e
    finally:
        if isinstance(error, KeyboardInterrupt):
            raise error
        return plst, error



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import networkx as nx
import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as ndi
import math

def quickVisual(G, showLabel = False):
    """Just makes a simple _matplotlib_ figure and displays it, with each node coloured by its type. You can add labels with _showLabel_. This looks a bit nicer than the one provided my _networkx_'s defaults.

    # Parameters

    _showLabel_ : `optional [bool]`

    > Default `False`, if `True` labels will be added to the nodes giving their IDs.
    """
    colours = "brcmykwg"
    f = plt.figure(1)
    ax = f.add_subplot(1,1,1)
    ndTypes = []
    ndColours = []
    layout = nx.spring_layout(G, k = 4 / math.sqrt(len(G.nodes())))
    for nd in G.nodes(data = True):
        if 'type' in nd[1]:
            if nd[1]['type'] not in ndTypes:
                ndTypes.append(nd[1]['type'])
            ndColours.append(colours[ndTypes.index(nd[1]['type']) % len(colours)])
        elif len(ndColours) > 1:
            raise RuntimeError("Some nodes do not have a type")
    if len(ndColours) < 1:
        nx.draw_networkx_nodes(G, pos = layout, node_color = colours[0], node_shape = '8', node_size = 100, ax = ax)
    else:
        nx.draw_networkx_nodes(G, pos = layout, node_color = ndColours, node_shape = '8', node_size = 100, ax = ax)
    nx.draw_networkx_edges(G, pos = layout, width = .7, ax = ax)
    if showLabel:
        nx.draw_networkx_labels(G, pos = layout, font_size = 8, ax = ax)
    plt.axis('off')
    f.set_facecolor('w')

def graphDensityContourPlot(G, iters = 50, layout = None, layoutScaleFactor = 1, overlay = False, nodeSize = 10, axisSamples = 100, blurringFactor = .1, contours = 15, graphType = 'coloured'):
    """Creates a 3D plot giving the density of nodes on a 2D plane, as a surface in 3D.

    Most of the options are for tweaking the final appearance. _layout_ and _layoutScaleFactor_ allow a pre-layout graph to be provided. If a layout is not provided the [networkx.spring_layout()](https://networkx.github.io/documentation/latest/reference/generated/networkx.drawing.layout.spring_layout.html) is used after _iters_ iterations. Then, once the graph has been laid out a grid of _axisSamples_ cells by _axisSamples_ cells is overlaid and the number of nodes in each cell is determined, a gaussian blur is then applied with a sigma of _blurringFactor_. This then forms a surface in 3 dimensions, which is then plotted.

    If you find the resultant image looks too banded raise the the _contours_ number to ~50.

    # Parameters

    _G_ : `networkx Graph`

    > The graph to be plotted

    _iters_ : `optional [int]`

    > Default `50`, the number of iterations for the spring layout if _layout_ is not provided.

    _layout_ : `optional [networkx layout dictionary]`

    > Default `None`, if provided will be used as a layout of the graph, the maximum distance from the origin along any axis must also given as _layoutScaleFactor_, which is by default `1`.

    _layoutScaleFactor_ : `optional [double]`

    > Default `1`, The maximum distance from the origin allowed along any axis given by _layout_, i.e. the layout must fit in a square centered at the origin with side lengths 2 * _layoutScaleFactor_

    _overlay_ : `optional [bool]`

    > Default `False`, if `True` the 2D graph will be plotted on the X-Y plane at Z = 0.

    _nodeSize_ : `optional [double]`

    > Default `10`, the size of the nodes dawn in the overlay

    _axisSamples_ : `optional [int]`

    > Default 100, the number of cells used along each axis for sampling. A larger number will mean a lower average density.

    _blurringFactor_ : `optional [double]`

    > Default `0.1`, the sigma value used for smoothing the surface density. The higher this number the smoother the surface.

    _contours_ : `optional [int]`

    > Default 15, the number of different heights drawn. If this number is low the resultant image will look very banded. It is recommended this be raised above `50` if you want your images to look good, **Warning** this will make them much slower to generate and interact with.

    _graphType_ : `optional [str]`

    > Default `'coloured'`, if `'coloured'` the image will have a destiny based colourization applied, the only other option is `'solid'` which removes the colourization.

    """
    from mpl_toolkits.mplot3d import Axes3D

    if not isinstance(G, nx.classes.digraph.DiGraph) and not isinstance(G, nx.classes.graph.Graph):
        raise TypeError("{} is not a valid input.".format(type(G)))
    if layout is None:
        layout = nx.spring_layout(G, scale = axisSamples - 1, iterations = iters)
        grid = np.zeros( [axisSamples, axisSamples],dtype=np.float32)
        for v in layout.values():
            x, y = tuple(int(x) for x in v.round(0))
            grid[y][x] += 1
    elif isinstance(layout, dict):
        layout = layout.copy()
        grid = np.zeros([axisSamples, axisSamples],dtype=np.float32)
        multFactor = (axisSamples - 1) / layoutScaleFactor
        for k in layout.keys():
            tmpPos = layout[k] * multFactor
            layout[k] = tmpPos
            x, y = tuple(int(x) for x in tmpPos.round(0))
            grid[y][x] += 1
    else:
        raise TypeError("{} is not a valid input.".format(type(layout)))
    fig = plt.figure()
    #axis = fig.add_subplot(111)
    axis = fig.gca(projection='3d')
    if overlay:
        nx.draw_networkx(G, pos = layout, ax = axis, node_size = nodeSize, with_labels = False, edgelist = [])
    grid = ndi.gaussian_filter(grid, (blurringFactor * axisSamples, blurringFactor * axisSamples))
    X = Y = np.arange(0, axisSamples, 1)
    X, Y = np.meshgrid(X, Y)
    if graphType == "solid":
        CS = axis.plot_surface(X,Y, grid)
    else:
        CS = axis.contourf(X, Y, grid, contours)
    axis.set_xlabel('X')
    axis.set_ylabel('Y')
    axis.set_zlabel('Node Density')



import csv
import os.path

from .baseGrant import Grant, csvAndLinesReader
from ..mkExceptions import BadGrant

class NSERCGrant(Grant):
    def __init__(self, original, grantdDict, sFile, sLine):
        bad = False
        error = None
        if grantdDict.get('Cle', '') == '':
            bad = True
            error = BadGrant("Missing 'CLE'")
            idValue = "NSERC:{}".format(hash(original))
        else:
            idValue = "NSERC:{}".format(grantdDict.get('Cle', ''))
        Grant.__init__(self, original, grantdDict, idValue, bad, error, sFile = sFile, sLine = sLine)

    def update(self, other):
        for field, value in other._fieldDict.items():
            if value == '':
                continue
            elif self._fieldDict.get(field, '') == '':
                self._fieldDict[field] = value
            else:
                self._fieldDict[field] += "; {}".format(value)

    def getInvestigators(self, tags = None, seperator = ";", _getTag = False):
        """Returns a list of the names of investigators. The optional arguments are ignored.

        # Returns

        `list [str]`

        > A list of all the found investigator's names
        """
        if tags is None:
            tags = []
        elif isinstance(tags, str):
            tags = [tags]
        for k in self.keys():
            if 'name-' in k.lower() and k not in tags:
                tags.append(k)
        return super().getInvestigators(tags = tags, seperator = seperator, _getTag = _getTag)

    def getInstitutions(self, tags = None, seperator = ";", _getTag = False):
        """Returns a list with the names of the institution. The optional arguments are ignored

        # Returns

        `list [str]`

        > A list with 1 entry the name of the institution
        """
        if tags is None:
            tags = []
        elif isinstance(tags, str):
            tags = [tags]
        for k in self.keys():
            if 'institution' in k.lower() and k not in tags:
                tags.append(k)
        return super().getInvestigators(tags = tags, seperator = seperator, _getTag = _getTag)

def isNSERCfile(fileName, useFileName = True):
    if useFileName and not os.path.basename(fileName).startswith('NSERC_'):
        return False
    try:
        with open(fileName, 'r', encoding = 'latin-1') as openfile:
            reader = csv.DictReader(openfile, fieldnames=None, dialect='excel')
            length = 0
            for row in reader:
                length += 1
                if set(row.keys()) != set(reader.fieldnames):
                    return False
            if length < 1:
                return False
    except (StopIteration, UnicodeDecodeError, KeyError):
        return False
    else:
        return True

def parserNSERCfile(fileName):
    grantSet = set()
    error = None
    try:
        with open(fileName, 'r', encoding = 'latin-1') as openfile:
            f = enumerate(openfile, start = 1)
            reader = csvAndLinesReader(f, fieldnames = None, dialect = 'excel')
            for lineNum, lineString, lineDict in reader:
                G = NSERCGrant(lineString, lineDict, sFile = fileName, sLine = lineNum)
                if G in grantSet:
                    for Gin in grantSet:
                        if Gin == G:
                            Gin.update(G)
                            break
                else:
                    grantSet.add(G)
    except Exception:
        if error is None:
            error = BadGrant("The file '{}' is having decoding issues. It may have been modifed since it was downloaded or not be a NSERC grant file.".format(fileName))
    except KeyboardInterrupt as e:
        error = e
    finally:
        if isinstance(error, KeyboardInterrupt):
            raise error
        return grantSet, error



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import os
import os.path
import csv
import re
try:
    import collections.abc
except ImportError:
    import collections
    collections.abc = collections
import networkx as nx

from .constants import __version__
from .mkRecord import Record, _pandasPrep
from .progressBar import _ProgressBar
from .WOS.tagProcessing.funcDicts import tagToFullDict, fullToTagDict, normalizeToTag
from .citation import Citation
from .fileHandlers import recordHandlers
from .mkExceptions import BadWOSRecord, RCTypeError, BadInputFile, BadRecord, RCValueError, RecordsNotCompatible, UnknownFile

from .mkCollection import CollectionWithIDs

from .scopus.scopusHandlers import scopusHeader

import metaknowledge

class RecordCollection(CollectionWithIDs):
    """A container for a large number of indivual records.

    `RecordCollection` provides ways of creating [Records](./Record.html#metaknowledge.Record) from an isi file, string, list of records or directory containing isi files.

    When being created if there are issues the Record collection will be declared bad, `bad` wil be set to `False`, it will then mostly return `None` or False. The attribute `error` contains the exception that occurred.

    They also possess an attribute `name` also accessed with `__repr__()`, this is used to auto generate the names of files and can be set at creation, note though that any operations that modify the RecordCollection's contents will update the name to include what occurred.

    # Customizations

    The Records are containing within a set and as such many of the set operations are defined, pop, union, in ... also records are hashed with their WOS string so no duplication can occur. The comparison operators `<`, `<=`, `>`, `>=` are based strictly on the number of Records within the collection, while equality looks for an exact match on the Records

    # \_\_Init\_\_

    _inCollection_ is the object containing the information about the Records to be constructed it can be an isi file, string, list of records or directory containing isi files

    # Parameters

    _inCollection_ : `optional [str] or None`

    > the name of the source of WOS records. It can be skipped to produce an empty collection.

    > If a file is provided. First it is checked to see if it is a WOS file (the header is checked). Then records are read from it one by one until the 'EF' string is found indicating the end of the file.

    > If a directory is provided. First each file in the directory is checked for the correct header and all those that do are then read like indivual files. The records are then collected into a single set in the RecordCollection.

    _name_ : `optional [str]`

    > The name of the RecordCollection, defaults to empty string. If left empty the name of the Record collection is set to the name of the file or directory used to create the collection. If provided the name id set to _name_

    _extension_ : `optional [str]`

    > The extension to search for when reading a directory for files. _extension_ is the suffix searched for when a directory is read for files, by default it is empty so all files are read.

    _cached_ : `optional [bool]`

    > Default `False`, if `True` and the _inCollection_ is a directory (a string giving the path to a directory) then the initialized `RecordCollection` will be saved in the directory as a Python pickle with the suffix `'.mkDirCache'`. Then if the `RecordCollection` is initialized a second time it will be recovered from the file, which is much faster than reprising every file in the directory.

    > _metaknowledge_ saves the names of the parsed files as well as their last modification times and will check these when recreating the `RecordCollection`, so modifying existing files or adding new ones will result in the entire directory being reanalyzed and a new cache file being created. The extension given to `__init__()` is taken into account as well and each suffix is given its own cache.

    > **Note** The pickle allows for arbitrary python code execution so only use caches that you trust.
    """

    def __init__(self, inCollection = None, name = '', extension = '', cached = False, quietStart = False):
        progArgs = (0, "Starting to make a RecordCollection")
        if metaknowledge.VERBOSE_MODE and not quietStart:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            bad = False
            errors = {}
            name = name
            recordTypes = set()
            if cached:
                cacheName = None
            if inCollection is None:
                PBar.updateVal(.5, "Empty RecordCollection created")
                if not name:
                    name = "Empty"
                recordsSet = set()
            elif isinstance(inCollection, str):
                inCollection = os.path.realpath(os.path.expanduser(inCollection))
                if os.path.isfile(inCollection):
                    PBar.updateVal(.2, "RecordCollection from a file started")
                    if not inCollection.endswith(extension):
                        raise RCTypeError("extension of input file does not match requested extension")
                    if not name:
                        name = os.path.splitext(os.path.split(inCollection)[1])[0]
                    try:
                        for recordType, processor, detector in recordHandlers:
                            if detector(inCollection):
                                recordTypes.add(recordType)
                                recordsSet, pError = processor(inCollection)
                                if pError is not None:
                                    bad = True
                                    errors[inCollection] = pError
                                break
                    except UnknownFile:
                        raise BadInputFile("'{}' does not match any known file type.\nIts header might be damaged or it could have been modified by another program.".format(inCollection))
                elif os.path.isdir(inCollection):
                    count = 0
                    PBar.updateVal(0, "RecordCollection from files in {}".format(inCollection))
                    if extension and not name:
                        name = "{}-files-from-{}".format(extension, inCollection)
                    elif not name:
                        name = "files-from-{}".format(inCollection)
                    recordsSet = set()
                    flist = []
                    for f in os.listdir(inCollection):
                        fullF = os.path.join(os.path.abspath(inCollection), f)
                        if fullF.endswith(extension) and not fullF.endswith('mkRecordDirCache') and os.path.isfile(fullF):
                            flist.append(fullF)
                    if cached:
                        PBar.updateVal(0, "Trying to load from cache")
                        cacheName = os.path.join(inCollection, '{}.[{}].mkRecordDirCache'.format(os.path.basename(os.path.abspath(inCollection)), extension))
                        if self._loadFromCache(cacheName, flist, name, extension):
                            try:
                                PBar.finish("Done reloading {} Records from cache".format(len(self)))
                            except AttributeError:
                                PBar.finish("Done reloading from the cache {}. Warning an error occured.".format(cacheName))
                            return
                        else:
                            PBar.updateVal(0, 'Cache error, rereading files')
                    for fileName in flist:
                        count += 1
                        PBar.updateVal(count / len(flist), "Reading records from: {}".format(fileName))
                        try:
                            for recordType, processor, detector in recordHandlers:
                                if detector(fileName):
                                    recordTypes.add(recordType)
                                    recs, pError = processor(fileName)
                                    if pError is not None:
                                        bad = True
                                        errors[fileName] = pError
                                    recordsSet |= recs
                                    break
                        except UnknownFile:
                            if extension != '':
                                raise BadInputFile("'{}' does not match any known file type, but has the requested extension '{}'. Its header might be damaged or it could have been modified by another program.".format(fileName, extension))
                            else:
                                pass
                else:
                    raise RCTypeError("'{}' is not a path to a directory or file. Strings cannot be used to initialize RecordCollections".format(inCollection))
            elif isinstance(inCollection, collections.abc.Iterable):
                PBar.updateVal(.5, "RecordCollection from an iterable started")
                for R in inCollection:
                    if not isinstance(R, Record):
                        raise RCTypeError("RecordCollections can only contain Records, '{}' is not a valid part of an input iterable.".format(R))
                recordsSet = set(inCollection)
            else:
                raise RCTypeError("A RecordCollection cannot be created from {}.".format(inCollection))
            CollectionWithIDs.__init__(self, recordsSet, Record, recordTypes, name, bad, errors)
            if cached:
                if cacheName is None:
                    pass
                else:
                    PBar.updateVal(1, "Writing RecordCollection cache to {}".format(cacheName))
                    self._createCache(cacheName, flist, name, extension)
            try:
                PBar.finish("Done making a RecordCollection of {} Records".format(len(self)))
            except AttributeError:
                PBar.finish("Done making a RecordCollection. Warning an error occured.")

    def __bytes__(self):
        encoding = self.peek().encoding()
        try:
            return bytes('\n', encoding = encoding).join((bytes(R) for R in self))
        except BadRecord as e:
            raise e from None

    def __add__(self, other):
        self_name = '' if not hasattr(self, 'name') else self.name
        other_name = '' if not hasattr(other, 'name') else other.name

        new_col = metaknowledge.RecordCollection(name='{} {}'.format(self_name, other_name))
        for rec in self:
            new_col.add(rec)

        for rec in other:
            new_col.add(rec)

        return new_col

    def dropNonJournals(self, ptVal = 'J', dropBad = True, invert = False):
        """Drops the non journal type `Records` from the collection, this is done by checking _ptVal_ against the PT tag

        # Parameters

        _ptVal_ : `optional [str]`

        > Default `'J'`, The value of the PT tag to be kept, default is `'J'` the journal tag, other tags can be substituted.

        _dropBad_ : `optional [bool]`

        > Default `True`, if `True` bad `Records` will be dropped as well those that are not journal entries

        _invert_ : `optional [bool]`

        > Default `False`, Set `True` to drop journals (or the PT tag given by _ptVal_) instead of keeping them. **Note**, it still drops bad Records if _dropBad_ is `True`
        """
        if dropBad:
            self.dropBadEntries()
        if invert:
            self._collection = {r for r in self._collection if r['pubType'] != ptVal.upper()}
        else:
            self._collection = {r for r in self._collection if r['pubType'] == ptVal.upper()}

    def writeFile(self, fname = None):
        """Writes the `RecordCollection` to a file, the written file's format is identical to those download from WOS. The order of `Records` written is random.

        # Parameters

        _fname_ : `optional [str]`

        > Default `None`, if given the output file will written to _fanme_, if `None` the `RecordCollection`'s name's first 200 characters are used with the suffix .isi
        """
        if len(self._collectedTypes) < 2:
            recEncoding = self.peek().encoding()
        else:
            recEncoding = 'utf-8'
        if fname:
            f = open(fname, mode = 'w', encoding = recEncoding)
        else:
            f = open(self.name[:200] + '.txt', mode = 'w', encoding = recEncoding)
        if self._collectedTypes == {'WOSRecord'}:
            f.write("\ufeffFN Thomson Reuters Web of Science\u2122\n")
            f.write("VR 1.0\n")
        elif self._collectedTypes == {'MedlineRecord'}:
            f.write('\n')
        elif self._collectedTypes == {'ScopusRecord'}:
            f.write("\ufeff{}\n".format(','.join(scopusHeader)))
        for R in self._collection:
            R.writeRecord(f)
            f.write('\n')
        if self._collectedTypes == {'WOSRecord'}:
            f.write('EF')
        f.close()

    def writeCSV(self, fname = None, splitByTag = None, onlyTheseTags = None, numAuthors = True, genderCounts = True, longNames = False, firstTags = None, csvDelimiter = ',', csvQuote = '"', listDelimiter = '|'):
        """Writes all the `Records` from the collection into a csv file with each row a record and each column a tag.

        # Parameters

        _fname_ : `optional [str]`

        > Default `None`, the name of the file to write to, if `None` it uses the collections name suffixed by .csv.

        _splitByTag_ : `optional [str]`

        > Default `None`, if a tag is given the output will be divided into different files according to the value of the tag, with only the records associated with that tag. For example if `'authorsFull'` is given then each file will only have the lines for `Records` that author is named in.

        > The file names are the values of the tag followed by a dash then the normale name for the file as given by _fname_, e.g. for the year 2016 the file could be called `'2016-fname.csv'`.

        _onlyTheseTags_ : `optional [iterable]`

        > Default `None`, if an iterable (list, tuple, etc) only the tags in _onlyTheseTags_ will be used, if not given then all tags in the records are given.

        > If you want to use all known tags pass [metaknowledge.knownTagsList](./ExtendedRecord.html#metaknowledge.ExtendedRecord.tagProcessingFunc).

        _numAuthors_ : `optional [bool]`

        > Default `True`, if `True` adds the number of authors as the column `'numAuthors'`.

        _longNames_ : `optional [bool]`

        > Default `False`, if `True` will convert the tags to their longer names, otherwise the short 2 character ones will be used.

        _firstTags_ : `optional [iterable]`

        > Default `None`, if `None` the iterable `['UT', 'PT', 'TI', 'AF', 'CR']` is used. The tags given by the iterable are the first ones in the csv in the order given.

        > **Note** if tags are in _firstTags_ but not in _onlyTheseTags_, _onlyTheseTags_ will override _firstTags_

        _csvDelimiter_ : `optional [str]`

        > Default `','`, the delimiter used for the cells of the csv file.

        _csvQuote_ : `optional [str]`

        > Default `'"'`, the quote character used for the csv.

        _listDelimiter_ : `optional [str]`

        > Default `'|'`, the delimiter used between values of the same cell if the tag for that record has multiple outputs.
        """
        if firstTags is None:
            firstTags = ['id', 'title', 'authorsFull', 'citations', 'keywords', 'DOI']
        for i in range(len(firstTags)):
            if firstTags[i] in fullToTagDict:
                firstTags[i] = fullToTagDict[firstTags[i]]
        if onlyTheseTags:
            for i in range(len(onlyTheseTags)):
                if onlyTheseTags[i] in fullToTagDict:
                    onlyTheseTags[i] = fullToTagDict[onlyTheseTags[i]]
            retrievedFields = [t for t in firstTags if t in onlyTheseTags] + [t for t in onlyTheseTags if t not in firstTags]
        else:
            retrievedFields = firstTags
            for R in self:
                tagsLst = [t for t in R.keys() if t not in retrievedFields]
                retrievedFields += tagsLst
        if longNames:
            try:
                retrievedFields = [tagToFullDict[t] for t in retrievedFields]
            except KeyError:
                raise KeyError("One of the tags could not be converted to a long name.")
        if fname:
            baseFileName = fname
        else:
            baseFileName = "{}.csv".format(self.name[:200])
        if numAuthors:
            csvWriterFields = retrievedFields + ["num-Authors"]
        else:
            csvWriterFields = retrievedFields
        if genderCounts:
            csvWriterFields += ['num-Male', 'num-Female', 'num-Unknown']
        if splitByTag is None:
            f = open(baseFileName, mode = 'w', encoding = 'utf-8', newline = '')
            csvWriter = csv.DictWriter(f, csvWriterFields, delimiter = csvDelimiter, quotechar = csvQuote, quoting=csv.QUOTE_ALL)
            csvWriter.writeheader()
        else:
            filesDict = {}
        for R in self:
            if splitByTag:
                try:
                    splitVal = R[splitByTag]
                except KeyError:
                    continue
                else:
                    if not isinstance(splitVal, list):
                        splitVal = [str(splitVal)]
            recDict = {}
            for t in retrievedFields:
                value = R.get(t)
                if isinstance(value, str):
                    recDict[t] = value
                elif hasattr(value, '__iter__'):
                    recDict[t] = listDelimiter.join([str(v) for v in value])
                elif value is None:
                    recDict[t] = ''
                else:
                    recDict[t] = str(value)
            if numAuthors:
                recDict["num-Authors"] = len(R.get('authorsShort', []))
            if genderCounts:
                recDict['num-Male'], recDict['num-Female'], recDict['num-Unknown'] = R.authGenders(_countsTuple = True)
            if splitByTag:
                for sTag in splitVal:
                    if sTag in filesDict:
                        filesDict[sTag][1].writerow(recDict)
                    else:
                        fname = "{}-{}".format(sTag[:200], baseFileName)
                        f = open(fname, mode = 'w', encoding = 'utf-8', newline = '')
                        csvWriter = csv.DictWriter(f, csvWriterFields, delimiter = csvDelimiter, quotechar = csvQuote, quoting=csv.QUOTE_ALL)
                        csvWriter.writeheader()
                        csvWriter.writerow(recDict)
                        filesDict[sTag] = (f, csvWriter)
            else:
                csvWriter.writerow(recDict)
        if splitByTag:
            for f, c in filesDict.values():
                f.close()
        else:
            f.close()

    def writeBib(self, fname = None, maxStringLength = 1000, wosMode = False, reducedOutput = False, niceIDs = True):
        """Writes a bibTex entry to _fname_ for each `Record` in the collection.

        If the Record is of a journal article (PT J) the bibtext type is set to `'article'`, otherwise it is set to `'misc'`. The ID of the entry is the WOS number and all the Record's fields are given as entries with their long names.

        **Note** This is not meant to be used directly with LaTeX none of the special characters have been escaped and there are a large number of unnecessary fields provided. _niceID_ and _maxLength_ have been provided to make conversions easier only.

        **Note** Record entries that are lists have their values separated with the string `' and '`, as this is the way bibTex understands

        # Parameters

        _fname_ : `optional [str]`

        > Default `None`, The name of the file to be written. If not given one will be derived from the collection and the file will be written to .

        _maxStringLength_ : `optional [int]`

        > Default 1000, The max length for a continuous string. Most bibTex implementation only allow string to be up to 1000 characters ([source](https://www.cs.arizona.edu/~collberg/Teaching/07.231/BibTeX/bibtex.html)), this splits them up into substrings then uses the native string concatenation (the `'#'` character) to allow for longer strings

        _WOSMode_ : `optional [bool]`

        > Default `False`, if `True` the data produced will be unprocessed and use double curly braces. This is the style WOS produces bib files in and mostly macthes that.

        _restrictedOutput_ : `optional [bool]`

        > Default `False`, if `True` the tags output will be limited to: `'AF'`, `'BF'`, `'ED'`, `'TI'`, `'SO'`, `'LA'`, `'NR'`, `'TC'`, `'Z9'`, `'PU'`, `'J9'`, `'PY'`, `'PD'`, `'VL'`, `'IS'`, `'SU'`, `'PG'`, `'DI'`, `'D2'`, and `'UT'`

        _niceID_ : `optional [bool]`

        > Default `True`, if `True` the IDs used will be derived from the authors, publishing date and title, if `False` it will be the UT tag
        """
        if fname:
            f = open(fname, mode = 'w', encoding = 'utf-8')
        else:
            f = open(self.name[:200] + '.bib', mode = 'w', encoding = 'utf-8')
        f.write("%This file was generated by the metaknowledge Python package.\n%The contents have been automatically generated and are likely to not work with\n%LaTeX without some human intervention. This file is meant for other automatic\n%systems and not to be used directly for making citations\n")
        #I figure this is worth mentioning, as someone will get annoyed at none of the special characters being escaped and how terrible some of the fields look to humans
        for R in self:
            try:
                f.write('\n\n')
                f.write(R.bibString(maxLength =  maxStringLength, WOSMode = wosMode, restrictedOutput = reducedOutput, niceID = niceIDs))
            except BadWOSRecord:
                pass
            except AttributeError:
                raise RecordsNotCompatible("The Record '{}', with ID '{}' does not support writing to bibtext files.".format(R, R.id))
        f.close()

    def findProbableCopyright(self):
        """Finds the (likely) copyright string from all abstracts in the `RecordCollection`

        # Returns

        `list[str]`

        > A deduplicated list of all the copyright strings
        """
        retCopyrights = set()
        for R in self:
            begin, abS = findCopyright(R.get('abstract', ''))
            if abS != '':
                retCopyrights.add(abS)
        return list(retCopyrights)

    def forBurst(self, tag, outputFile = None, dropList = None, lower = True, removeNumbers = True, removeNonWords = True, removeWhitespace = True, stemmer = None):
        """Creates a pandas friendly dictionary with 2 columns one `'year'` and the other `'word'`. Each row is a word that occurred in the field given by _tag_ in a `Record` and the year of the record. Unfortunately getting the month or day with any type of accuracy has proved to be impossible so year is the only option.

        # Parameters

        _tag_ : `str`

        > The tag giving the field for the words to be extracted from.

        _outputFile_ : `optional str`

        > Default `None`, if a path is given a csv file will be created from the returned dictionary and written to that file

        _dropList_ : `optional list[str]`

        > Default `None`, if a list of strings is given each field will be checked for substrings, before any other processing, in the field, surrounded by spaces, matching those in _dropList_. The strings will only be dropped if they are surrounded on both sides with spaces (`' '`) so if `dropList = ['a']` then `'a cat'` will become `'cat'`.

        _lower_ : `optional bool`

        > default `True`, if `True` the output will made lower case

        _removeNumbers_ : `optional bool`

        > default `True`, if `True` all numbers will be removed

        _removeNonWords_ : `optional bool`

        > default `True`, if `True` all non-number non-number characters will be removed

        _removeWhitespace_ : `optional bool`

        > default `True`, if `True` all whitespace will be converted to a single space (`' '`)

        _stemmer_ : `optional func`

        > default `None`, if a function is provided it will be run on each individual word in the field and the output will replace it. For example to use the  `PorterStemmer` in the _nltk_ package you would give `nltk.PorterStemmer().stem`
        """

        whiteSpaceRegex = re.compile(r'\s+')

        if removeNumbers:
            if removeNonWords:
                otherString = r"[\W\d]"
            else:
                otherString = r"\d"
        elif removeNonWords:
            otherString = r"\W"
        else:
            otherString = ''

        def otherRepl(r):
            if r.group(0) == ' ':
                return ' '
            else:
                return ''
        otherDropsRegex = re.compile(otherString)

        def burstPreper(inString):
            if dropList is not None:
                inString = " {} ".format(inString)
                for dropS in (" {} ".format(s) for s in dropList):
                    if dropS in inString:
                        inString = inString.replace(dropS, ' ')
                inString = inString[1:-1]
            if removeWhitespace:
                inString = re.sub(whiteSpaceRegex, lambda x: ' ', inString, count = 0)
            if lower:
                inString = inString.lower()
            inString = re.sub(otherDropsRegex, otherRepl, inString, count = 0)
            sTokens = inString.split(' ')
            if stemmer is not None:
                retTokens = []
                for token in sTokens:
                    if stemmer is not None:
                        token = stemmer(token)
                    retTokens.append(token)
            else:
                retTokens = sTokens
            return retTokens

        retDict = {'year' : [], 'word' : []}

        pcount = 0
        pmax = len(self)
        progArgs = (0, "Starting to work on DataFrame for burst analysis")
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            for R in self:
                pcount += 1
                PBar.updateVal(pcount/ pmax, "Analyzing: {}".format(R))
                try:
                    year = R['year']
                except KeyError:
                    continue
                try:
                    burstVal = R[tag]
                except KeyError:
                    continue
                else:
                    if isinstance(burstVal, list):
                        burstVal = ' '.join((str(i) for i in burstVal))
                    else:
                        burstVal = str(burstVal)
                for sToken in burstPreper(burstVal):
                    retDict['year'].append(year)
                    retDict['word'].append(sToken)

            if outputFile is not None:
                PBar.updateVal(.99, "Writing to file: {}".format(outputFile))
                with open(outputFile, 'w', newline = '') as f:
                    writer = csv.DictWriter(f, ['year', 'word'])
                    for row in range(len(retDict['year'])):
                        writer.writerow({k : retDict[k][row] for k in retDict.keys()})
            PBar.finish("Done burst analysis DataFrame with {} rows".format(len(retDict['year'])))
        return retDict

    def forNLP(self, outputFile = None, extraColumns = None, dropList = None, lower = True, removeNumbers = True, removeNonWords = True, removeWhitespace = True, removeCopyright = False, stemmer = None):
        """Creates a pandas friendly dictionary with each row a `Record` in the `RecordCollection` and the columns fields natural language processing uses (id, title, publication year, keywords and the abstract). The abstract is by default is processed to remove non-word, non-space characters and the case is lowered.

        # Parameters

        _outputFile_ : `optional str`

        > default `None`, if a file path is given a csv of the returned data will be written

        _extraColumns_ : `optional list[str]`

        > default `None`, if a list of tags is given each of the tag's values for a `Record` will be added to the output(s)

        _dropList_ : `optional list[str]`

        > default `None`, if a list of strings is provided they will be dropped from the output's abstracts. The matching is case sensitive and done before any other processing. The strings will only be dropped if they are surrounded on both sides with spaces (`' '`) so if `dropList = ['a']` then `'a cat'` will become `'cat'`.

        _lower_ : `optional bool`

        > default `True`, if `True` the abstract will made to lower case

        _removeNumbers_ : `optional bool`

        > default `True`, if `True` all numbers will be removed

        _removeNonWords_ : `optional bool`

        > default `True`, if `True` all non-number non-number characters will be removed

        _removeWhitespace_ : `optional bool`

        > default `True`, if `True` all whitespace will be converted to a single space (`' '`)

        _removeCopyright_ : `optional bool`

        > default `False`, if `True` the copyright statement at the end of the abstract will be removed and added to a new column. Note this is heuristic based and will not work for all papers.

        _stemmer_ : `optional func`

        > default `None`, if a function is provided it will be run on each individual word in the abstract and the output will replace it. For example to use the  `PorterStemmer` in the _nltk_ package you would give `nltk.PorterStemmer().stem`
        """
        whiteSpaceRegex = re.compile(r'\s+')

        if removeNumbers:
            if removeNonWords:
                otherString = r"[\W\d]"
            else:
                otherString = r"\d"
        elif removeNonWords:
            otherString = r"\W"
        else:
            otherString = ''

        def otherRepl(r):
            if r.group(0) == ' ':
                return ' '
            else:
                return ''
        otherDropsRegex = re.compile(otherString)

        def abPrep(abst):
            if dropList is not None:
                #incase a drop string is on the edge
                abst = " {} ".format(abst.replace('\n', ' '))
                for dropS in (" {} ".format(s) for s in dropList):
                    if dropS in abst:
                        abst = abst.replace(dropS, ' ')
                abst = abst[1:-1]
            if removeWhitespace:
                abst = re.sub(whiteSpaceRegex, lambda x: ' ', abst, count = 0)
            if removeCopyright:
                abst, copyrightString = findCopyright(abst)
            else:
                copyrightString = ''
            if lower:
                abst = abst.lower()
            abst = re.sub(otherDropsRegex, otherRepl, abst, count = 0)
            if stemmer is not None:
                sTokens = abst.split(' ')
                retTokens = []
                for token in sTokens:
                    if stemmer is not None:
                        token = stemmer(token)
                    retTokens.append(token)
                abst = ' '.join(retTokens)
            return abst, copyrightString

        pcount = 0
        pmax = len(self)
        progArgs = (0, "Starting to work on DataFrame for NLP")
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}

        retDict = {'id' : [], 'year' : [], 'title' : [], 'keywords' : [], 'abstract' : []}
        if removeCopyright:
            retDict['copyright'] = []
        if extraColumns is None:
            extraColumns = []
        else:
            for builtinColumn in ['id', 'year', 'title', 'keywords', 'abstract']:
                if builtinColumn in extraColumns:
                    extraColumns.remove(builtinColumn)
        for column in extraColumns:
            retDict[column] = []
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            for R in self:
                pcount += 1
                PBar.updateVal(pcount/ pmax, "Analyzing: {}".format(R))
                abstract, copyrightString = abPrep(R.get('AB', ''))

                retDict['id'].append(R.id)
                retDict['year'].append(R.get('year', ''))
                retDict['title'].append(R.get('title', ''))
                retDict['keywords'].append('|'.join(R.get('keywords', [])))
                retDict['abstract'].append(abstract)
                if removeCopyright:
                    retDict['copyright'].append(copyrightString)
                for extraTag in extraColumns:
                    e = R.get(extraTag)
                    if isinstance(e, list):
                        e = '|'.join((str(s) for s in e))
                    elif e is None:
                        e = ''
                    retDict[extraTag].append(e)

            if outputFile is not None:
                PBar.updateVal(.99, "Writing to file: {}".format(outputFile))
                with open(outputFile, 'w', newline = '') as f:
                    fieldNames = list(retDict.keys())
                    fieldNames.remove('id')
                    fieldNames.remove('title')
                    fieldNames.remove('year')
                    fieldNames.remove('keywords')
                    fieldNames = ['id', 'year', 'title', 'keywords'] + fieldNames
                    writer = csv.DictWriter(f, fieldNames)
                    writer.writeheader()
                    for row in range(len(retDict['id'])):
                        writer.writerow({k : retDict[k][row] for k in retDict.keys()})
            PBar.finish("Done NLP DataFrame with {} rows".format(len(retDict['id'])))
        return retDict

    def makeDict(self, onlyTheseTags = None, longNames = False, raw = False, numAuthors = True, genderCounts = True):
        """Returns a dict with each key a tag and the values being lists of the values for each of the Records in the collection, `None` is given when there is no value and they are in the same order across each tag.

        When used with pandas: `pandas.DataFrame(RC.makeDict())` returns a data frame with each column a tag and each row a Record.

        # Parameters

        _onlyTheseTags_ : `optional [iterable]`

        > Default `None`, if an iterable (list, tuple, etc) only the tags in _onlyTheseTags_ will be used, if not given then all tags in the records are given.

        > If you want to use all known tags pass [metaknowledge.knownTagsList](./ExtendedRecord.html#metaknowledge.ExtendedRecord.tagProcessingFunc).

        _longNames_ : `optional [bool]`

        > Default `False`, if `True` will convert the tags to their longer names, otherwise the short 2 character ones will be used.

        _cleanedVal_ : `optional [bool]`

        > Default `True`, if `True` the processed values for each `Record`'s field will be provided, otherwise the raw values are given.

        _numAuthors_ : `optional [bool]`

        > Default `True`, if `True` adds the number of authors as the column `'numAuthors'`.
        """
        if onlyTheseTags:
            for i in range(len(onlyTheseTags)):
                if onlyTheseTags[i] in fullToTagDict:
                    onlyTheseTags[i] = fullToTagDict[onlyTheseTags[i]]
            retrievedFields = onlyTheseTags
        else:
            retrievedFields = []
            for R in self:
                tagsLst = [t for t in R.keys() if t not in retrievedFields]
                retrievedFields += tagsLst
        if longNames:
            try:
                retrievedFields = [tagToFullDict[t] for t in retrievedFields]
            except KeyError:
                raise KeyError("One of the tags could not be converted to a long name.")
        retDict = {k : [] for k in retrievedFields}
        if numAuthors:
            retDict["num-Authors"] = []
        if genderCounts:
            retDict.update({'num-Male' : [], 'num-Female' : [], 'num-Unknown' : []})
        for R in self:
            if numAuthors:
                retDict["num-Authors"].append(len(R.get('authorsShort', [])))
            if genderCounts:
                m, f, u = R.authGenders(_countsTuple = True)
                retDict['num-Male'].append(m)
                retDict['num-Female'].append(f)
                retDict['num-Unknown'].append(u)
            for k, v in R.subDict(retrievedFields, raw = raw).items():
                retDict[k].append(v)
        return retDict

    def rpys(self, minYear = None, maxYear = None, dropYears = None, rankEmptyYears = False):
        """This implements _Referenced Publication Years Spectroscopy_ a techinique for finding import years in citation data. The authors of the original papers have a website with more information, found [here](http://www.leydesdorff.net/software/rpys/).

        This function computes the spectra of the `RecordCollection` and returns a dictionary mapping strings to lists of `ints`. Each list is ordered and the values of each with the same index form a row and each list a column. The strings are the names of the columns. This is intended to be read directly by pandas `DataFrames`.

        The columns returned are:

        1. `'year'`, the years of the counted citations, missing years are inserted with a count of 0, unless they are outside the bounds of the highest year or the lowest year and the default value is used. e.g. if the highest year is 2016, 2017 will not be inserted unless _maxYear_ has been set to 2017 or higher
        2. `'count'`, the number of times the year was cited
        3. `'abs-deviation'`, deviation from the 5-year median. Calculated by taking the absolute deviation of the count from the median of it and the next 2 years and the preceding 2 years
        4. `'rank'`, the rank of the year, the highest ranked year being the one with the highest deviation, the second highest being the second highest deviation and so on. All years with 0 count are given the rank 0 by default

        # Parameters

        _minYear_ : `optional int`

        > Default `1000`, The lowest year to be returned, note years outside this bound will be used to calculate the deviation from the 5-year median

        _maxYear_ : `optional int`

        > Default `2100`, The highest year to be returned, note years outside this bound will be used to calculate the deviation from the 5-year median

        _dropYears_ : `optional int or list[int]`

        > Default `None`, year or collection of years that will be removed from the returned value, note the dropped years will still be used to calculate the deviation from the 5-year

        _rankEmptyYears_ : `optional [bool]`

        > Default `False`, if `True` years with 0 count will be ranked according to their deviance, if many 0 count years exist their ordering is not guaranteed to be stable

        # Returns

        `dict[str:list]`

        > The table of values from the _Referenced Publication Years Spectroscopy_
        """

        def deviation(targetYear, targetValue, targetDict):
            yearCounts = [targetValue]
            for deltaY in [-2, -1, 1, 2]:
                try:
                    yearCounts.append(targetDict[targetYear + deltaY])
                except KeyError:
                    yearCounts.append(0)
            medianCount = list(sorted(yearCounts))[2]
            absDiff = targetValue - medianCount
            return absDiff

        if dropYears is None:
            dropYears = set()
        yearCounts = {}
        retDict = {'year' : [], 'count' : [], 'abs-deviation' : [], 'rank' : []}

        for R in self:
            try:
                cites = R['citations']
            except KeyError:
                continue
            recYear = R.get('year', float('inf'))
            for cite in cites:
                try:
                    #year can be None
                    cYear = int(cite.year)
                except (AttributeError, TypeError):
                    continue
                else:
                    #need the extra years for the normlization
                    if (maxYear is not None and cYear > (maxYear + 2)) or (minYear is not None and cYear < (minYear - 2)):
                        continue
                    #years from before the paper are an error
                    elif recYear < (cYear + 2):
                        continue
                if cYear in yearCounts:
                    yearCounts[cYear] += 1
                else:
                    yearCounts[cYear] = 1

        if minYear is None:
            smallest = min(yearCounts.keys())
            if smallest > 1000:
                minYear = smallest
        if maxYear is None:
            biggest = max(yearCounts.keys())
            if biggest < 2100:
                maxYear = biggest

        targetYears = set(( i for i in range(minYear, maxYear + 1) if i not in dropYears))

        ranks = {}
        yearDeviances = {}

        for y in targetYears:
            try:
                c = yearCounts[y]
            except KeyError:
                c = 0
            yearDeviances[y] = deviation(y, c, yearCounts)

        for rank, year in enumerate(sorted(yearDeviances.items(), key = lambda x: x[1], reverse = False), start = 1):
            ranks[year[0]] = rank

        for y in targetYears:
            try:
                c = yearCounts[y]
            except KeyError:
                c = 0
            if c == 0 and not rankEmptyYears:
                retDict['rank'].append(0)
            else:
                retDict['rank'].append(ranks[y])
            retDict['abs-deviation'].append(yearDeviances[y])
            retDict['year'].append(y)
            retDict['count'].append(c)

        return retDict

    def genderStats(self, asFractions = False):
        """Creates a dict (`{'Male' : maleCount, 'Female' : femaleCount, 'Unknown' : unknownCount}`) with the numbers of male, female and unknown names in the collection.

        # Parameters

        _asFractions_ : `optional bool`

        > Default `False`, if `True` the counts will be divided by the total number of names, giving the fraction of names in each category instead of the raw counts.

        # Returns

        `dict[str:int]`

        > A dict with three keys `'Male'`, `'Female'` and `'Unknown'` mapping to their respective counts
        """

        maleCount = 0
        femaleCount = 0
        unknownCount = 0
        for R in self:
            m, f, u = R.authGenders(_countsTuple = True)
            maleCount += m
            femaleCount += f
            unknownCount += u
        if asFractions:
            tot = maleCount + femaleCount + unknownCount
            return {'Male' : maleCount / tot, 'Female' : femaleCount / tot, 'Unknown' : unknownCount / tot}
        return {'Male' : maleCount, 'Female' : femaleCount, 'Unknown' : unknownCount}

    def getCitations(self, field = None, values = None, pandasFriendly = True, counts = True):
        """Creates a pandas ready dict with each row a different citation the contained Records and columns containing the original string, year, journal, author's name and the number of times it occured.

        There are also options to filter the output citations with _field_ and _values_

        # Parameters

        _field_ : `optional str`

        > Default `None`, if given all citations missing the named field will be dropped.

        _values_ : `optional str or list[str]`

        > Default `None`, if _field_ is also given only those citations with one of the strings given in _values_ will be included.

        > e.g. to get only citations from 1990 or 1991: `field = year, values = [1991, 1990]`

        _pandasFriendly_ : `optional bool`

        > Default `True`, if `False` a list of the citations will be returned instead of the more complicated pandas dict

        _counts_ : `optional bool`

        > Default `True`, if `False` the counts columns will be removed

        # Returns

        `dict`

        > A pandas ready dict with all the Citations
        """
        retCites = []
        if values is not None:
            if isinstance(values, (str, int, float)) or not isinstance(values, collections.abc.Container):
                values = [values]
        for R in self:
            retCites += R.getCitations(field = field, values = values, pandasFriendly = False)
        if pandasFriendly:
            return _pandasPrep(retCites, counts)
        else:
            return list(set(retCites))

    def networkCoAuthor(self, detailedInfo = False, weighted = True, dropNonJournals = False, count = True, useShortNames = False, citeProfile = False):
        """Creates a coauthorship network for the RecordCollection.

        # Parameters

        _detailedInfo_ : `optional [bool or iterable[WOS tag Strings]]`

        > Default `False`, if `True` all nodes will be given info strings composed of information from the Record objects themselves. This is Equivalent to passing the list: `['PY', 'TI', 'SO', 'VL', 'BP']`.

        > If _detailedInfo_ is an iterable (that evaluates to `True`) of WOS Tags (or long names) The values  of those tags will be used to make the info attributes.

        > For each of the selected tags an attribute will be added to the node using the values of those tags on the first `Record` encountered. **Warning** iterating over `RecordCollection` objects is not deterministic the first `Record` will not always be same between runs. The node will be given attributes with the names of the WOS tags for each of the selected tags. The attributes will contain strings of containing the values (with commas removed), if multiple values are encountered they will be comma separated.

        > Note: _detailedInfo_ is not identical to the _detailedCore_ argument of [Recordcollection.networkCoCitation()](#metaknowledge.RecordCollection.networkCoCitation) or [Recordcollection.networkCitation()](#metaknowledge.RecordCollection.networkCitation)

        _weighted_ : `optional [bool]`

        > Default `True`, whether the edges are weighted. If `True` the edges are weighted by the number of co-authorships.

        _dropNonJournals_ : `optional [bool]`

        > Default `False`, whether to drop authors from non-journals

        _count_ : `optional [bool]`

        > Default `True`, causes the number of occurrences of a node to be counted

        # Returns

        `Networkx Graph`

        > A networkx graph with author names as nodes and collaborations as edges.
        """
        grph = nx.Graph()
        pcount = 0
        progArgs = (0, "Starting to make a co-authorship network")
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        if bool(detailedInfo):
            try:
                infoVals = []
                for tag in detailedInfo:
                    infoVals.append(normalizeToTag(tag))
            except TypeError:
                infoVals = ['year', 'title', 'journal', 'volume', 'beginningPage']
            def attributeMaker(Rec):
                attribsDict = {}
                for val in infoVals:
                    recVal = Rec.get(val)
                    if isinstance(recVal, list):
                        attribsDict[val] = ', '.join((str(v).replace(',', '') for v in recVal))
                    else:
                        attribsDict[val] = str(recVal).replace(',', '')
                if count:
                    attribsDict['count'] = 1
                if citeProfile:
                    attribsDict['citeProfile'] = {}
                return attribsDict
        else:
            if count:
                if citeProfile:
                    attributeMaker = lambda x: {'count' : 1, 'citeProfile' : {}}
                else:
                    attributeMaker = lambda x: {'count' : 1}
            else:
                if citeProfile:
                    attributeMaker = lambda x: {'citeProfile' : {}}
                else:
                    attributeMaker = lambda x: {}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            for R in self:
                if PBar:
                    pcount += 1
                    PBar.updateVal(pcount/ len(self), "Analyzing: " + str(R))
                if dropNonJournals and not R.createCitation().isJournal():
                    continue
                if useShortNames:
                    authsList = R.get('authorsShort', [])
                else:
                    authsList = R.get('authorsFull', [])
                if authsList:
                    authsList = list(authsList)
                    detailedInfo = attributeMaker(R)
                    if citeProfile:
                        citesLst = R.get('citations', [])
                    for i, auth1 in enumerate(authsList):
                        if auth1 not in grph:
                            grph.add_node(auth1, **detailedInfo.copy())
                        elif count:
                            grph.nodes[auth1]['count'] += 1
                        if citeProfile:
                            for c in citesLst:
                                try:
                                    grph.nodes[auth1]['citeProfile'][c] += 1
                                except KeyError:
                                    grph.nodes[auth1]['citeProfile'][c] = 1
                        for auth2 in authsList[i + 1:]:
                            if auth2 not in grph:
                                grph.add_node(auth2, **detailedInfo.copy())
                            elif count:
                                grph.nodes[auth2]['count'] += 1
                            if citeProfile:
                                for c in citesLst:
                                    try:
                                        grph.nodes[auth2]['citeProfile'][c] += 1
                                    except KeyError:
                                        grph.nodes[auth2]['citeProfile'][c] = 1
                            if grph.has_edge(auth1, auth2) and weighted:
                                grph.edges[auth1, auth2]['weight'] += 1
                            elif weighted:
                                grph.add_edge(auth1, auth2, weight = 1)
                            else:
                                grph.add_edge(auth1, auth2)
            if citeProfile:
                if PBar:
                    PBar.updateVal(.99, "Extracting citation profiles")
                previous = {}
                for n, dat in grph.nodes(data = True):
                    previous[n] = dat
                    #zip(*l) undoes zip(l1, l2)
                    try:
                        cites, counts = zip(*dat['citeProfile'].items())
                    except ValueError:
                        cites, counts = [], []
                    dat['citeProfileCites'] = '|'.join((str(c) for c in cites))
                    dat['citeProfileCounts'] = '|'.join((str(c) for c in counts))
                    del dat['citeProfile']
            if PBar:
                PBar.finish("Done making a co-authorship network from {}".format(self))
        return grph

    def networkCoCitation(self, dropAnon = True, nodeType = "full", nodeInfo = True, fullInfo = False, weighted = True, dropNonJournals = False, count = True, keyWords = None, detailedCore = True, detailedCoreAttributes = False, coreOnly = False, expandedCore = False, addCR = False):
        """Creates a co-citation network for the RecordCollection.

        # Parameters

        _nodeType_ : `optional [str]`

        > One of `"full"`, `"original"`, `"author"`, `"journal"` or `"year"`. Specifies the value of the nodes in the graph. The default `"full"` causes the citations to be compared holistically using the [metaknowledge.Citation](./Citation.html#metaknowledge.citation.Citation) builtin comparison operators. `"original"` uses the raw original strings of the citations. While `"author"`, `"journal"` and `"year"` each use the author, journal and year respectively.

        _dropAnon_ : `optional [bool]`

        > default `True`, if `True` citations labeled anonymous are removed from the network

        _nodeInfo_ : `optional [bool]`

        > default `True`, if `True` an extra piece of information is stored with each node. The extra inforamtion is detemined by _nodeType_.

        _fullInfo_ : `optional [bool]`

        > default `False`, if `True` the original citation string is added to the node as an extra value, the attribute is labeled as fullCite

        _weighted_ : `optional [bool]`

        > default `True`, wether the edges are weighted. If `True` the edges are weighted by the number of citations.

        _dropNonJournals_ : `optional [bool]`

        > default `False`, wether to drop citations of non-journals

        _count_ : `optional [bool]`

        > default `True`, causes the number of occurrences of a node to be counted

        _keyWords_ : `optional [str] or [list[str]]`

        > A string or list of strings that the citations are checked against, if they contain any of the strings they are removed from the network

        _detailedCore_ : `optional [bool or iterable[WOS tag Strings]]`

        > default `True`, if `True` all Citations from the core (those of records in the RecordCollection) and the _nodeType_ is `'full'` all nodes from the core will be given info strings composed of information from the Record objects themselves. This is Equivalent to passing the list: `['AF', 'PY', 'TI', 'SO', 'VL', 'BP']`.

        > If _detailedCore_ is an iterable (That evaluates to `True`) of WOS Tags (or long names) The values  of those tags will be used to make the info attribute. All

        > The resultant string is the values of each tag, with commas removed, seperated by `', '`, just like the info given by non-core Citations. Note that for tags like `'AF'` that return lists only the first entry in the list will be used. Also a second attribute is created for all nodes called inCore wich is a boolean describing if the node is in the core or not.

        > Note: _detailedCore_  is not identical to the _detailedInfo_ argument of [Recordcollection.networkCoAuthor()](#metaknowledge.RecordCollection.networkCoAuthor)

        _coreOnly_ : `optional [bool]`

        > default `False`, if `True` only Citations from the RecordCollection will be included in the network

        _expandedCore_ : `optional [bool]`

        > default `False`, if `True` all citations in the ouput graph that are records in the collection will be duplicated for each author. If the nodes are `"full"`, `"original"` or `"author"` this will result in new noded being created for the other options the results are **not** defined or tested. Edges will be created between each of the nodes for each record expanded, attributes will be copied from exiting nodes.

        # Returns

        `Networkx Graph`

        > A networkx graph with hashes as ID and co-citation as edges
        """
        allowedTypes = ["full", "original", "author", "journal", "year"]
        if nodeType not in allowedTypes:
            raise RCValueError("{} is not an allowed nodeType.".format(nodeType))
        coreValues = []
        if bool(detailedCore):
            try:
                for tag in detailedCore:
                    coreValues.append(normalizeToTag(tag))
            except TypeError:
                coreValues = ['id', 'authorsFull', 'year', 'title', 'journal', 'volume', 'beginningPage']
        tmpgrph = nx.Graph()
        pcount = 0
        progArgs = (0, "Starting to make a co-citation network")
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            if coreOnly or coreValues or expandedCore:
                coreCitesDict = {R.createCitation() : R for R in self}
                if coreOnly:
                    coreCites = coreCitesDict.keys()
                else:
                    coreCites = None
            else:
                coreCitesDict = None
                coreCites = None
            for R in self:
                if PBar:
                    pcount += 1
                    PBar.updateVal(pcount / len(self), "Analyzing: {}".format(R))
                Cites = R.get('citations')
                if Cites:
                    filteredCites = filterCites(Cites, nodeType, dropAnon, dropNonJournals, keyWords, coreCites)
                    addToNetwork(tmpgrph, filteredCites, count, weighted, nodeType, nodeInfo , fullInfo, coreCitesDict, coreValues, detailedCoreAttributes, addCR, headNd = None)
            if expandedCore:
                if PBar:
                    PBar.updateVal(.98, "Expanding core Records")
                expandRecs(tmpgrph, self, nodeType, weighted)
            if PBar:
                PBar.finish("Done making a co-citation network from {}".format(self))
        return tmpgrph

    def networkCitation(self, dropAnon = False, nodeType = "full", nodeInfo = True, fullInfo = False, weighted = True, dropNonJournals = False, count = True, directed = True, keyWords = None, detailedCore = True, detailedCoreAttributes = False, coreOnly = False, expandedCore = False, recordToCite = True, addCR = False, _quiet = False):
        """Creates a citation network for the RecordCollection.

        # Parameters

        _nodeType_ : `optional [str]`

        > One of `"full"`, `"original"`, `"author"`, `"journal"` or `"year"`. Specifies the value of the nodes in the graph. The default `"full"` causes the citations to be compared holistically using the [metaknowledge.Citation](./Citation.html#metaknowledge.citation.Citation) builtin comparison operators. `"original"` uses the raw original strings of the citations. While `"author"`, `"journal"` and `"year"` each use the author, journal and year respectively.

        _dropAnon_ : `optional [bool]`

        > default `True`, if `True` citations labeled anonymous are removed from the network

        _nodeInfo_ : `optional [bool]`

        > default `True`, whether an extra piece of information is stored with each node.

        _fullInfo_ : `optional [bool]`

        > default `False`, whether the original citation string is added to the node as an extra value, the attribute is labeled as fullCite

        _weighted_ : `optional [bool]`

        > default `True`, whether the edges are weighted. If `True` the edges are weighted by the number of citations.

        _dropNonJournals_ : `optional [bool]`

        > default `False`, whether to drop citations of non-journals

        _count_ : `optional [bool]`

        > default `True`, causes the number of occurrences of a node to be counted

        _keyWords_ : `optional [str] or [list[str]]`

        > A string or list of strings that the citations are checked against, if they contain any of the strings they are removed from the network

        _directed_ : `optional [bool]`

        > Determines if the output graph is directed, default `True`

        _detailedCore_ : `optional [bool or iterable[WOS tag Strings]]`

        > default `True`, if `True` all Citations from the core (those of records in the RecordCollection) and the _nodeType_ is `'full'` all nodes from the core will be given info strings composed of information from the Record objects themselves. This is Equivalent to passing the list: `['AF', 'PY', 'TI', 'SO', 'VL', 'BP']`.

        > If _detailedCore_ is an iterable (That evaluates to `True`) of WOS Tags (or long names) The values  of those tags will be used to make the info attribute. All

        > The resultant string is the values of each tag, with commas removed, seperated by `', '`, just like the info given by non-core Citations. Note that for tags like `'AF'` that return lists only the first entry in the list will be used. Also a second attribute is created for all nodes called inCore wich is a boolean describing if the node is in the core or not.

        > Note: _detailedCore_  is not identical to the _detailedInfo_ argument of [Recordcollection.networkCoAuthor()](#metaknowledge.RecordCollection.networkCoAuthor)

        _coreOnly_ : `optional [bool]`

        > default `False`, if `True` only Citations from the RecordCollection will be included in the network

        _expandedCore_ : `optional [bool]`

        > default `False`, if `True` all citations in the ouput graph that are records in the collection will be duplicated for each author. If the nodes are `"full"`, `"original"` or `"author"` this will result in new noded being created for the other options the results are **not** defined or tested. Edges will be created between each of the nodes for each record expanded, attributes will be copied from exiting nodes.

        # Returns

        `Networkx DiGraph or Networkx Graph`

        > See _directed_ for explanation of returned type

        > A networkx digraph with hashes as ID and citations as edges
        """
        allowedTypes = ["full", "original", "author", "journal", "year"]
        if nodeType not in allowedTypes:
            raise RCValueError("{} is not an allowed nodeType.".format(nodeType))
        coreValues = []
        if bool(detailedCore):
            try:
                for tag in detailedCore:
                    coreValues.append(normalizeToTag(tag))
            except TypeError:
                coreValues = ['id', 'authorsFull', 'year', 'title', 'journal', 'volume', 'beginningPage']
        if directed:
            tmpgrph = nx.DiGraph()
        else:
            tmpgrph = nx.Graph()
        pcount = 0
        progArgs = (0, "Starting to make a citation network")
        if metaknowledge.VERBOSE_MODE and not _quiet:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            if coreOnly or coreValues:
                coreCitesDict = {R.createCitation() : R for R in self}
                if coreOnly:
                    coreCites = coreCitesDict.keys()
                else:
                    coreCites = None
            else:
                coreCitesDict = None
                coreCites = None
            for R in self:
                if PBar:
                    pcount += 1
                    PBar.updateVal(pcount/ len(self), "Analyzing: " + str(R))
                reRef = R.createCitation()
                if len(filterCites([reRef], nodeType, dropAnon, dropNonJournals, keyWords, coreCites)) == 0:
                    continue
                rCites = R.get('citations')
                if rCites:
                    filteredCites = filterCites(rCites, nodeType, dropAnon, dropNonJournals, keyWords, coreCites)
                    addToNetwork(tmpgrph, filteredCites, count, weighted, nodeType, nodeInfo, fullInfo, coreCitesDict, coreValues, detailedCoreAttributes, addCR, recordToCite, headNd = reRef)
            if expandedCore:
                if PBar:
                    PBar.updateVal(.98, "Expanding core Records")
                expandRecs(tmpgrph, self, nodeType, weighted)
            PBar.finish("Done making a citation network from {}".format(self))
        return tmpgrph

    def networkBibCoupling(self, weighted = True, fullInfo = False, addCR = False):
        """Creates a bibliographic coupling network based on citations for the RecordCollection.

        # Parameters

        _weighted_ : `optional bool`

        > Default `True`, if `True` the weight of the edges will be added to the network

        _fullInfo_ : `optional bool`

        > Default `False`, if `True` the full citation string will be added to each of the nodes of the network.

        # Returns

        `Networkx Graph`

        > A graph of the bibliographic coupling
        """
        progArgs = (0, "Make a citation network for coupling")
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            citeGrph = self.networkCitation(weighted = False, directed = True, detailedCore = True, fullInfo = fullInfo, count = False, nodeInfo = True, addCR = addCR, _quiet = True)
            pcount = 0
            pmax = len(citeGrph)
            PBar.updateVal(.2, "Starting to classify nodes")
            workingGrph = nx.Graph()
            couplingSet = set()
            for n, d in citeGrph.nodes(data = True):
                pcount += 1
                PBar.updateVal(.2 + .4 * (pcount / pmax), "Classifying: {}".format(n))
                if d['inCore']:
                    workingGrph.add_node(n, **d)
                if citeGrph.in_degree(n) > 0:
                    couplingSet.add(n)
            pcount = 0
            pmax = len(couplingSet)
            for n in couplingSet:
                PBar.updateVal(.6 + .4 * (pcount / pmax), "Coupling: {}".format(n))
                citesLst = list(citeGrph.in_edges(n))
                for i, edgeOuter in enumerate(citesLst):
                    outerNode = edgeOuter[0]
                    for edgeInner in citesLst[i + 1:]:
                        innerNode = edgeInner[0]
                        if weighted and  workingGrph.has_edge(outerNode, innerNode):
                            workingGrph.edges[outerNode, innerNode]['weight'] += 1
                        elif weighted:
                            workingGrph.add_edge(outerNode, innerNode, weight = 1)
                        else:
                            workingGrph.add_edge(outerNode, innerNode)
            PBar.finish("Done making a bib-coupling network from {}".format(self))
        return workingGrph

    def _extractTagged(self, taglist):
        recordsWithTags = set()
        for R in self:
            for t in taglist:
                hasTags = True
                if t not in R.tags:
                    hasTags = False
                    break
            if hasTags:
                recordsWithTags.add(R)
        return RecordCollection(recordsWithTags, repr(self) + "_tags(" + ','.join(taglist) + ')', quietStart = True)

    def yearSplit(self, startYear, endYear, dropMissingYears = True):
        """Creates a RecordCollection of Records from the years between _startYear_ and _endYear_ inclusive.

        # Parameters

        _startYear_ : `int`

        > The smallest year to be included in the returned RecordCollection

        _endYear_ : `int`

        > The largest year to be included in the returned RecordCollection

        _dropMissingYears_ : `optional [bool]`

        > Default `True`, if `True` Records with missing years will be dropped. If `False` a `TypeError` exception will be raised

        # Returns

        `RecordCollection`

        > A RecordCollection of Records from _startYear_ to _endYear_
        """
        recordsInRange = set()
        for R in self:
            try:
                if R.get('year') >= startYear and R.get('year') <= endYear:
                    recordsInRange.add(R)
            except TypeError:
                if dropMissingYears:
                    pass
                else:
                    raise
        RCret = RecordCollection(recordsInRange, name = "{}({}-{})".format(self.name, startYear, endYear), quietStart = True)
        RCret._collectedTypes = self._collectedTypes.copy()
        return RCret

    def localCiteStats(self, pandasFriendly = False, keyType = "citation"):
        """Returns a dict with all the citations in the CR field as keys and the number of times they occur as the values

        # Parameters

        _pandasFriendly_ : `optional [bool]`

        > default `False`, makes the output be a dict with two keys one `'Citations'` is the citations the other is their occurrence counts as `'Counts'`.

        _keyType_ : `optional [str]`

        > default `'citation'`, the type of key to use for the dictionary, the valid strings are `'citation'`, `'journal'`, `'year'` or `'author'`. IF changed from `'citation'` all citations matching the requested option will be contracted and their counts added together.

        # Returns

        `dict[str, int or Citation : int]`

        > A dictionary with keys as given by _keyType_ and integers giving their rates of occurrence in the collection
        """
        count = 0
        recCount = len(self)
        progArgs = (0, "Starting to get the local stats on {}s.".format(keyType))
        if metaknowledge.VERBOSE_MODE:
            progKwargs = {'dummy' : False}
        else:
            progKwargs = {'dummy' : True}
        with _ProgressBar(*progArgs, **progKwargs) as PBar:
            keyTypesLst = ["citation", "journal", "year", "author"]
            citesDict = {}
            if keyType not in keyTypesLst:
                raise TypeError("{} is not a valid key type, only '{}' or '{}' are.".format(keyType, "', '".join(keyTypesLst[:-1]), keyTypesLst[-1]))
            for R in self:
                rCites = R.get('citations')
                if PBar:
                    count += 1
                    PBar.updateVal(count / recCount, "Analysing: {}".format(R.UT))
                if rCites:
                    for c in rCites:
                        if keyType == keyTypesLst[0]:
                            cVal = c
                        else:
                            cVal = getattr(c, keyType)
                            if cVal is None:
                                continue
                        if cVal in citesDict:
                            citesDict[cVal] += 1
                        else:
                            citesDict[cVal] = 1
            if PBar:
                PBar.finish("Done, {} {} fields analysed".format(len(citesDict), keyType))
        if pandasFriendly:
            citeLst = []
            countLst = []
            for cite, occ in citesDict.items():
                citeLst.append(cite)
                countLst.append(occ)
            return {"Citations" : citeLst, "Counts" : countLst}
        else:
            return citesDict

    def localCitesOf(self, rec):
        """Takes in a Record, WOS string, citation string or Citation and returns a RecordCollection of all records that cite it.

        # Parameters

        _rec_ : `Record, str or Citation`

        > The object that is being cited

        # Returns

        `RecordCollection`

        > A `RecordCollection` containing only those `Records` that cite _rec_
        """
        localCites = []
        if isinstance(rec, Record):
            recCite = rec.createCitation()
        if isinstance(rec, str):
            try:
                recCite = self.getID(rec)
            except ValueError:
                try:
                    recCite = Citation(rec)
                except AttributeError:
                    raise ValueError("{} is not a valid WOS string or a valid citation string".format(recCite))
            else:
                if recCite is None:
                    return RecordCollection(inCollection = localCites, name = "Records_citing_{}".format(rec), quietStart = True)
                else:
                    recCite = recCite.createCitation()
        elif isinstance(rec, Citation):
            recCite = rec
        else:
            raise ValueError("{} is not a valid input, rec must be a Record, string or Citation object.".format(rec))
        for R in self:
            rCites = R.get('citations')
            if rCites:
                for cite in rCites:
                    if recCite == cite:
                        localCites.append(R)
                        break
        return RecordCollection(inCollection = localCites, name = "Records_citing_'{}'".format(rec), quietStart = True)

    def citeFilter(self, keyString = '', field = 'all', reverse = False, caseSensitive = False):
        """Filters `Records` by some string, _keyString_, in their citations and returns all `Records` with at least one citation possessing _keyString_ in the field given by _field_.

        # Parameters

        _keyString_ : `optional [str]`

        > Default `''`, gives the string to be searched for, if it is is blank then all citations with the specified field will be matched

        _field_ : `optional [str]`

        > Default `'all'`, gives the component of the citation to be looked at, it can be one of a few strings. The default is `'all'` which will cause the entire original `Citation` to be searched. It can be used to search across fields, e.g. `'1970, V2'` is a valid keystring
        The other options are:

        + `'author'`, searches the author field
        + `'year'`, searches the year field
        + `'journal'`, searches the journal field
        + `'V'`, searches the volume field
        + `'P'`, searches the page field
        + `'misc'`, searches all the remaining uncategorized information
        + `'anonymous'`, searches for anonymous `Citations`, _keyString_ is not ignored
        + `'bad'`, searches for bad citations, keyString is not used

        _reverse_ : `optional [bool]`

        > Default `False`, being set to `True` causes all `Records` not matching the query to be returned

        _caseSensitive_ : `optional [bool]`

        > Default `False`, if `True` causes the search across the original to be case sensitive, **only** the `'all'` option can be case sensitive
        """
        retRecs = []
        keyString = str(keyString)
        for R in self:
            try:
                if field == 'all':
                    for cite in R.get('citations'):
                        if caseSensitive:
                            if keyString in cite.original:
                                retRecs.append(R)
                                break
                        else:
                            if keyString.upper() in cite.original.upper():
                                retRecs.append(R)
                                break
                elif field == 'author':
                    for cite in R.get('citations'):
                        try:
                            if keyString.upper() in cite.author.upper():
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'journal':
                    for cite in R.get('citations'):
                        try:
                            if keyString.upper() in cite.journal:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'year':
                    for cite in R.get('citations'):
                        try:
                            if int(keyString) == cite.year:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'V':
                    for cite in R.get('citations'):
                        try:
                            if keyString.upper() in cite.V:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'P':
                    for cite in R.get('citations'):
                        try:
                            if keyString.upper() in cite.P:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'misc':
                    for cite in R.get('citations'):
                        try:
                            if keyString.upper() in cite.misc:
                                retRecs.append(R)
                                break
                        except AttributeError:
                            pass
                elif field == 'anonymous':
                    for cite in R.get('citations'):
                        if cite.isAnonymous():
                            retRecs.append(R)
                            break
                elif field == 'bad':
                    for cite in R.get('citations'):
                        if cite.bad:
                            retRecs.append(R)
                            break
            except TypeError:
                pass
        if reverse:
            excluded = []
            for R in self:
                if R not in retRecs:
                    excluded.append(R)
            return RecordCollection(inCollection = excluded, name = self.name, quietStart = True)
        else:
            return RecordCollection(inCollection = retRecs, name = self.name, quietStart = True)


def addToNetwork(grph, nds, count, weighted, nodeType, nodeInfo, fullInfo, coreCitesDict, coreValues, detailedValues, addCR, recordToCite = True, headNd = None):
    """Addeds the citations _nds_ to _grph_, according to the rules give by _nodeType_, _fullInfo_, etc.

    _headNd_ is the citation of the Record
    """
    if headNd is not None:
        hID = makeID(headNd, nodeType)
        if nodeType == 'full' or nodeType == 'original':
            hYear = getattr(headNd, "year")
        if hID not in grph:
            nodeName, nodeDat = makeNodeTuple(headNd, hID, nodeInfo, fullInfo, nodeType, count, coreCitesDict, coreValues, detailedValues, addCR)
            grph.add_node(nodeName, **nodeDat)
    else:
        hID = None
    idList = []
    yearList = []
    for n in nds:
        nID = makeID(n, nodeType)
        if nodeType == 'full' or nodeType == 'original':
            try:
                nYear = getattr(n, "year")
            except:
                nYear = None
            yearList.append(nYear)

        if nID not in grph:
            nodeName, nodeDat = makeNodeTuple(n, nID, nodeInfo, fullInfo, nodeType, count, coreCitesDict, coreValues, detailedValues, addCR)
            grph.add_node(nodeName, **nodeDat)
        elif count:
            grph.nodes[nID]['count'] += 1
        idList.append(nID)

    addedEdges = []
    if hID:
        for i in range(len(idList)):
            nID = idList[i]
            if nodeType == 'full' or nodeType == 'original':
                nYear = yearList[i]
                try:
                    yearDiff = abs(hYear - nYear)
                except:
                    yearDiff = None

                if weighted:
                    try:
                        if recordToCite:
                            grph[hID][nID]['weight'] += 1
                        else:
                            grph[nID][hID]['weight'] += 1
                    except KeyError:
                        if recordToCite:
                            grph.add_edge(hID, nID, weight=1, yearDiff=yearDiff)
                        else:
                            grph.add_edge(nID, hID, weight=1, yearDiff=yearDiff)
                elif nID not in grph[hID]:
                    addedEdges.append((hID, nID))

            elif weighted:
                try:
                    if recordToCite:
                        grph[hID][nID]['weight'] += 1
                    else:
                        grph[nID][hID]['weight'] += 1
                except KeyError:
                    if recordToCite:
                        grph.add_edge(hID, nID, weight=1)
                    else:
                        grph.add_edge(hID, nID, weight=1)
            elif nID not in grph[hID]:
                addedEdges.append((hID, nID, {yearDiff: yearDiff}))
    elif len(idList) > 1:
        for i, outerID in enumerate(idList):
            for innerID in idList[i + 1:]:
                if weighted:
                    try:
                        grph[outerID][innerID]['weight'] += 1
                    except KeyError:
                        grph.add_edge(outerID, innerID, weight = 1)
                elif innerID not in grph[outerID]:
                    addedEdges.append((outerID, innerID))
    grph.add_edges_from(addedEdges)


def makeID(citation, nodeType):
    """Makes the id, of the correct type for the network"""
    if nodeType != "full":
        return getattr(citation, nodeType)
    else:
        return citation.ID()


def makeNodeTuple(citation, idVal, nodeInfo, fullInfo, nodeType, count, coreCitesDict, coreValues, detailedValues, addCR):
    """Makes a tuple of idVal and a dict of the selected attributes"""
    d = {}
    if nodeInfo:
        if nodeType == 'full':
            if coreValues:
                if citation in coreCitesDict:
                    R = coreCitesDict[citation]
                    d['MK-ID'] = R.id
                    if not detailedValues:
                        infoVals = []
                        for tag in coreValues:
                            tagVal = R.get(tag)
                            if isinstance(tagVal, str):
                                infoVals.append(tagVal.replace(',',''))
                            elif isinstance(tagVal, list):
                                infoVals.append(tagVal[0].replace(',',''))
                            else:
                                pass
                        d['info'] = ', '.join(infoVals)
                    else:
                        for tag in coreValues:
                            v = R.get(tag, None)
                            if isinstance(v, list):
                                d[tag] = '|'.join(sorted(v))
                            else:
                                d[tag] = v
                    d['inCore'] = True
                    if addCR:
                        d['citations'] = '|'.join((str(c) for c in R.get('citations', [])))
                else:
                    d['MK-ID'] = 'None'
                    d['info'] = citation.allButDOI()
                    d['inCore'] = False
                    if addCR:
                        d['citations'] = ''
            else:
                d['info'] = citation.allButDOI()
        elif nodeType == 'journal':
            if citation.isJournal():
                d['info'] = str(citation.FullJournalName())
            else:
                d['info'] = "None"
        elif nodeType == 'original':
            d['info'] = str(citation)
        else:
            d['info'] = idVal
    if fullInfo:
        d['fullCite'] = str(citation)
    if count:
        d['count'] = 1
    return (idVal, d)


def filterCites(cites, nodeType, dropAnon, dropNonJournals, keyWords, coreCites):
    filteredCites = []
    for c in cites:
        if nodeType != "full" and hasattr(c, nodeType) and not getattr(c, nodeType):
            pass
        elif dropNonJournals and not c.isJournal():
            pass
        elif dropAnon and c.isAnonymous():
            pass
        elif coreCites is not None and c not in coreCites:
            pass
        elif keyWords:
            found = False
            citeString = str(c).upper()
            if isinstance(keyWords, str):
                if keyWords.upper() in citeString:
                    found = True
            else:
                for k in keyWords:
                    if k.upper() in citeString:
                        found = True
                        break
            if not found:
                filteredCites.append(c)
        else:
            filteredCites.append(c)
    return filteredCites


def expandRecs(G, RecCollect, nodeType, weighted):
    """Expand all the citations from _RecCollect_"""
    for Rec in RecCollect:
        fullCiteList = [makeID(c, nodeType) for c in Rec.createCitation(multiCite = True)]
        if len(fullCiteList) > 1:
            for i, citeID1 in enumerate(fullCiteList):
                if citeID1 in G:
                    for citeID2 in fullCiteList[i + 1:]:
                        if citeID2 not in G:
                            G.add_node(citeID2, **G.nodes[citeID1])
                            if weighted:
                                G.add_edge(citeID1, citeID2, weight = 1)
                            else:
                                G.add_edge(citeID1, citeID2)
                        elif weighted:
                            try:
                                G.edges[citeID1, citeID2]['weight'] += 1
                            except KeyError:
                                G.add_edge(citeID1, citeID2, weight = 1)
                        for e1, e2, data in G.edges(citeID1, data = True):
                            G.add_edge(citeID2, e2, **data)


def findCopyright(inS):
    possibleHits = ['. &COPY; ', '. Crown Copyright',' Elsevier Ltd. ', '. Copyright', '. Published by Els', '. (c) ', '. (C) ']
    splitString = False
    for target in possibleHits:
        if target in inS:
            splitString = target
            break
    if not splitString:
        regexHit = re.search(r'([.] [(]?(c|C)[)]? \d\d\d\d )|([.] \d\d\d\d Elsevier)', inS)
        if regexHit:
            splitString = regexHit.group(0)
    if splitString:
        copSplit = inS.split(splitString)
        if '.' == splitString[0]:
            return splitString.join(copSplit[:-1]) + '.', splitString[1:] + copSplit[-1].rstrip()
        else:
            return splitString.join(copSplit[:-1]), splitString + copSplit[-1].rstrip()
    else:
        return inS, ''



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
"""These are the functions used to process medline (pubmed) files at the backend. They are meant for use internal use by metaknowledge."""

from .tagProcessing.tagFunctions import *
from .tagProcessing.funcDicts import tagToFullDict, fullToTagDict, tagNameConverterDict, tagsAndNameSet, knownTagsList

from .recordWOS import WOSRecord, recordParser
from .wosHandlers import isWOSFile, wosParser



from ...WOS.tagProcessing.helpFuncs import getMonth

import re

def year(R):
    try:
        return int(R['DP'].split(' ')[0])
    except ValueError:
        yVal = re.search(r'-?\d{1,4}', R['DP'].split(' ')[0])
        if yVal is None:
            return 0
        else:
            return(int(yVal.group(0)))

def month(R):
    try:
        m = R['DP'].split(' ')[1]
    except IndexError:
        raise KeyError("Unable to extract a month")
    else:
        return getMonth(m)

def volume(R):
    """Returns the first number/word of the volume field, hopefully trimming something like: `'49 Suppl 20'` to `49`"""
    return R['VI'].split(' ')[0]

def beginningPage(R):
    """As pages may not be given as numbers this is the most accurate this function can be"""
    p = R['PG']
    if p.startswith('suppl '):
        p = p[6:]
    return p.split(' ')[0].split('-')[0].replace(';', '')


def DOI(R):
    ids = R['AID']
    for a in ids:
        if a.endswith(' [doi]'):
            return a[:-6]
    raise KeyError("No DOI number found")

def address(R):
    """Gets the first address of the first author"""
    return R['AD'][R['AU'][0]][0]

medlineSpecialTagToFunc = {
    'year' : year,
    'month' : month,
    'volume' : volume,
    'beginningPage' : beginningPage,
    'DOI' : DOI,
    'address' : address,

    'j9' : lambda R : R['TA'], #remaps to the closests field TA, but J9 != TA

    #'citations' : lambda R: None, #Medline does not have citations

    'grants' : lambda R: R['GR'],#This is the basis for the 'grants' special function

    'selfCitation' : lambda R: R.createCitation(), #just remaps to the correct function
    'authorsShort' : lambda R: R['AU'], #just remaps to the correct name
    'authorsFull' : lambda R : R['FAU'], #just remaps to the correct name
    'title' : lambda R : R['TI'], #just remaps to the correct name
    'journal' : lambda R : R['JT'], #just remaps to the correct name
    'keywords' : lambda R : R['OT'], #just remaps to the correct name
    'abstract' : lambda R : R['AB'], #just remaps to the correct name
    'id' : lambda R : R.id, #just remaps to the correct name
}



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import metaknowledge
import metaknowledge.journalAbbreviations
import networkx as nx
import argparse
import os
import sys
import readline
import collections
import code
import platform

#TODO: Figure output name stuff

#These are used when the user goes to the interactive terminal, so they must be global
RC = None
G = None

def argumentParser():
    parser = argparse.ArgumentParser(description="metaknowledge's simple command line interface")
    parser.add_argument("files", default = [], nargs = '*')
    parser.add_argument("--verbose", "-v", action = 'store_true' , default = False, help = "Verbose mode, every step is printed")
    parser.add_argument("--name", "-n",  default = False, help = "The name used for the recordCollection and resulting files.")
    parser.add_argument("--debug", "-d", action = 'store_true', default = False, help = "Enables debug messages.")
    parser.add_argument("--quiet", "-q", action = 'store_true', default = False, help = "Disables the progress bar. Progress bars only works on some operating system, if it does not work it will be disabled")
    parser.add_argument("--suffix", "-s", default = '', help = "The suffix of the WOS files you wish to extract Records from, by default all files are used and those that do not have Records are skipped")
    return parser.parse_args()

def yesorNo(prompt):
    while True:
        val = input(prompt).upper()
        if val == 'Y' or val == '':
            return True
        elif val == 'N':
            return False
        else:
            pass

def inputMenu(inputDict, header = None, footer = None, errorMsg = 'That is not an option please select a different value.', promptMsg = 'What is your selection: ', extraOptions = True):
    s = ''
    if header:
        s += '{}\n'.format(header)
    for k in inputDict.keys():
        s += '{0}) {1}\n'.format(k,inputDict[k])
    if extraOptions:
        s += 'i) open python console\n'
        s += 'q) quit\n'
    if footer:
        s += '{}\n'.format(footer)
    print(s, end = '')
    selection = input(promptMsg)
    if selection in inputDict:
        return selection
    elif extraOptions and selection == 'q':
        sys.exit()
    elif extraOptions and selection == 'i':
        bannerMsg = 'Python {0} on {1}\nType "help", "copyright", "credits" or "license" for more information.\nmetaknowledge is imported as "metaknowledge" and Networkx is imported as "nx".\n'.format(sys.version, platform.system())
        if RC:
            bannerMsg += "The metaknowledge RecordCollection is called 'RC'"
            if G:
                bannerMsg += " and the Networkx graph is called 'G'."
            else:
                bannerMsg += '.'
        code.interact(local = globals(), banner = bannerMsg)
    else:
        print('\033[93m\a' + errorMsg + '\033[0m') #Not windows compatible
        return inputMenu(inputDict, header = header, footer = footer, errorMsg = errorMsg, promptMsg = promptMsg)

def getOutputName(clargs, suffix, prompt = "What do you wish to call the output file? ", errorMsg = 'That is not a valid file name.', checking = True):
    if clargs.name:
        return clargs.name + suffix
    s  = input(prompt) + suffix
    if len(s) == 0:
        print(errorMsg)
        return getOutputName(clargs, suffix, prompt = prompt, errorMsg = errorMsg)
    else:
        if os.path.lexists(s) and checking:
            overWrite = yesorNo("{} already exists overwrite (y/n)? ".format(s))
            if overWrite:
                return s
            else:
                return getOutputName(clargs, suffix, prompt = prompt, errorMsg = errorMsg)
        else:
            return s

def Tag(prompt, collection, nMode = False):
    retTag = input(prompt).upper()
    if retTag in collection.tags():
        return retTag
    else:
        if nMode and retTag == '':
            return False
        else:
            print("{} is not a tag in any of the Records, please try again".format(retTag))
            return Tag(prompt, collection, nMode = nMode)

def getNum(prompt):
    retNum = input(prompt)
    try:
        return float(retNum)
    except ValueError:
        print("{} is not a number, please try again".format(retNum))
        return getNum(prompt)

def getFiles(args):
    tmpRC = metaknowledge.RecordCollection(name = '')
    if len(args.files) > 0:
        for f in args.files:
            path = os.path.abspath(os.path.expanduser(f))
            if os.path.exists(path):
                tmpRC |= metaknowledge.RecordCollection(path, extension = args.suffix)
            else:
                raise TypeError(path + " is not an existing file or directory")
        if args.name:
            tmpRC._repr = args.name
        else:
            FileNames = [os.path.basename(nm) for nm in args.files]
            if len(FileNames) == 1:
                tmpRC._repr = "WOS files from: {}".format(FileNames[0])
            else:
                tmpRC._repr = "WOS files from: {0} and {1}".format(', '.join(FileNames[:-1]), FileNames[-1])
        return tmpRC
    else:
        nflist = input("What files or directorys do you want to extract a network from: ")
        paths = [os.path.abspath(os.path.expanduser(f)) for f in nflist.split(' ')]
        badPaths = [f for f in paths if not os.path.exists(f)]
        if len(badPaths) > 1:
            print(', '.join(badPaths[:-1]) + " and " + badPaths[-1] + " are not existing files or directorys")
            args.files = []
            return getFiles(args)
        elif len(badPaths) == 1:
            print(badPaths[0] + " is not an existing file or directory")
            args.files = []
            return getFiles(args)
        else:
            args.files = paths
            return getFiles(args)

def getWhatToDo(clargs, inRC):
    promptDict = collections.OrderedDict([
    ('1', "Make a graph"),
    ('2', "Write the collection as a single WOS style file"),
    ('3', "Write the collection as a single WOS style file and make a graph"),
    ('4', "Write the collection as a single csv file"),
    ('5', "Write the collection as a single csv file and make a graph"),
    ('6', "Write all the citations to a single file"),
    ('7', "Go over non-journal citations"),
    ])
    print("A collection of {0} WOS records has been created".format(len(inRC)))
    optionNum = int(inputMenu(promptDict, header = "What do you wish to do with it:"))
    if optionNum == 1:
        return True
    elif optionNum == 2 or optionNum == 3:
        fName = getOutputName(clargs, '.txt')
        print("Writing {}".format(fName))
        inRC.writeFile(fname = fName)
        if optionNum == 2:
            return False
        else:
            return True
    elif optionNum == 4 or optionNum == 5:
        fName = getOutputName(clargs, '.csv')
        print("Writing {}".format(fName))
        inRC.writeCSV(fname = fName)
        if optionNum == 4:
            return False
        else:
            return True
    elif optionNum == 6:
        drop = yesorNo("Write only journals (y/n)? ")
        cites = []
        for R in inRC:
            if drop:
                cites += [str(c) + '\n' for c in R.get('citations', []) if c.isJournal()]
            else:
                cites += [str(c) + '\n' for c in R.get('citations', [])]
        fName = getOutputName(clargs, '.csv')
        cites = set(cites)
        print("Writing {}".format(fName))
        with open(fName, 'w') as f:
            f.writelines(cites)
        return False
    else:
        dbName = input("The default manual databse file is called {}, press Enter to use it or type the name of the database you wish to use:\n".format(metaknowledge.journalAbbreviations.manualDBname))
        print("Starting to go over citations, to exit press ctr-C.")
        if dbName == '':
            dbName = metaknowledge.journalAbbreviations.manualDBname
        try:
            for R in inRC:
                for c in R.get('citations', []):
                    if not hasattr(c, 'journal'):
                        print("{} does not have a journal field".format(c))
                    elif c.isJournal(manualDB= dbName, returnDict ='both'):
                        print("the journal field of {} is in the database".format(c.journal))
                    elif c.isJournal(manualDB= dbName, returnDict ='both', checkIfExcluded = True):
                        print("the journal field of {} is in the database marked to be skipped".format(c.journal))
                    else:
                        addToDB = yesorNo("The citation {} has the journal field:\n{} add as a journal (y/n)? ".format(c, c.journal))
                        if addToDB:
                            c.addToDB(manualDB= dbName)
                            print("{} added as a journal abbrviation.".format(c.journal))
                        else:
                            c.addToDB(invert = True)
                            print("{} will now be skipped".format(c.journal))
        except KeyboardInterrupt:
            print('\nExiting manual citation clasification', end = '')
            raise
        else:
            print('Done manual citation clasification')
            return False



def getNetwork(clargs, inRC):
    netsDict = collections.OrderedDict([
    ('1', "a one-mode network"),
    ('2', "a two-mode network"),
    ('3', "an n-mode network"),
    ('4', "a citation network"),
    ('5', "a co-citation network"),
    ('6', "a co-authorship network"),
    ])
    netID = int(inputMenu(netsDict, header = "What type of network do you wish to create?", promptMsg = "Input the number corresponding to the type of network you wish to generate? "))
    if netID == 1:
        otg = Tag("What is the tag to use for the network? ", inRC)
        print("Generating a network using the {0} tag.".format(otg), inRC)
        return inRC.networkOneMode(otg)
    elif netID == 2:
        tg1 = Tag("What is the first tag to use for the network? ", inRC)
        tg2 = Tag("And the second tag? ", inRC)
        print("Generating a network using the {0} and {1} tags.".format(tg1, tg2))
        return inRC.networkTwoMode(tg1, tg2)
    elif netID == 3:
        tgs = []
        tgs.append(Tag("What is the first tag to use for the network? ", inRC))
        innertag = Tag("And the next tag (leave blank to continue)? ", inRC, nMode = True)
        while innertag:
            tgs.append(innertag)
            innertag = Tag("And the next tag (leave blank to continue)? ", inRC, nMode = True)
        print("Generating a network using the {0} and {1} tags".format(', '.join(tgs[:-1]), tgs[-1]))
        return inRC.networkMultiMode(tgs)
    elif netID == 4:
        print("Generating citation network")
        return inRC.networkCitation()
    elif netID == 5:
        print("Generating co-citation network")
        return inRC.networkCoCitation()
    else:
        print("Generating co-authorship network")
        return inRC.networkCoAuthor()

def getThresholds(clargs, grph):
    thresDict = collections.OrderedDict([
    ('0', "Continue"),
    ('1', "Drop isolates"),
    ('2', "Remove self loops"),
    ('3', "Remove edges below some weight"),
    ('4', "Remove edges above some weight"),
    ('5', "Remove nodes below some degree"),
    ('6', "Remove nodes above some degree"),
    ])
    print("The network contains {0} nodes and {1} edges, of which {2} are isolated and {3} are self loops.".format(len(list(grph.nodes())), len(list(grph.edges())), len(list(nx.isolates(grph))), len(list(nx.selfloop_edges(grph)))))
    thresID = int(inputMenu(thresDict, header = "What type of filtering to you want? "))
    if thresID == 0:
        return grph
    elif thresID == 1:
        metaknowledge.dropNodesByDegree(grph, minDegree = 1)
        return getThresholds(clargs, grph)
    elif thresID == 2:
        metaknowledge.dropEdges(grph, dropSelfLoops = True)
        return getThresholds(clargs, grph)
    elif thresID == 3:
        metaknowledge.dropEdges(grph, minWeight = getNum("What is the minumum weight for an edge to be included? "))
        return getThresholds(clargs, grph)
    elif thresID == 4:
        metaknowledge.dropEdges(grph, minWeight = getNum("What is the maximum weight for an edge to be included? "))
        return getThresholds(clargs, grph)
    elif thresID == 5:
        metaknowledge.dropNodesByDegree(grph, minDegree = getNum("What is the minumum degree for an edge to be included? "))
        return getThresholds(clargs, grph)
    else:
        metaknowledge.dropNodesByDegree(grph, minDegree = getNum("What is the maximum degree for an edge to be included? "))
        return getThresholds(clargs, grph)

def  outputNetwork(clargs, grph):
    outDict = collections.OrderedDict([
    ('1', "edge list and node attribute list"),
    ('2', "edge list"),
    ('3', "node attribute list"),
    ('4', "graphml (SLOW)"),
    ])
    try:
        import metaknowledge.contour
    except ImportError:
        import metaknowledge
    else:
        outDict['0'] = "view graph"
        outDict.move_to_end('0', last = False)
    print("The network contains {} nodes and {} edges.".format(len(grph.nodes()), len(grph.edges())))
    outID = int(inputMenu(outDict, header = "What type of output to you want? "))
    if outID == 0:
        metaknowledge.contour.quickVisual(grph)
        outputNetwork(clargs, grph)
    elif outID == 1:
        while True:
            try:
                outName = getOutputName(clargs, '', checking = False)
                metaknowledge.writeGraph(grph, outName, overwrite = False)
            except OSError:
                if clargs.name:
                    metaknowledge.writeGraph(grph, outName, overwrite = True)
                    break
                else:
                    overWrite = yesorNo("{}, overwrite (y/n)? ")
                    if overWrite:
                        metaknowledge.writeGraph(grph, outName, overwrite = True)
                        break
                    else:
                        pass
            else:
                break

    elif outID == 2:
        outName = getOutputName(clargs, '.csv')
        metaknowledge.writeEdgeList(grph, outName)
    elif outID == 3:
        outName = getOutputName(clargs, '.csv')
        metaknowledge.writeNodeAttributeFile(grph, outName)
    else:
        outName = getOutputName(clargs, '.graphml')
        nx.write_graphml(grph, outName)

def mkCLI():
    try:
        args = argumentParser()
        if args.quiet:
            metaknowledge.VERBOSE_MODE = False
        global RC
        RC = getFiles(args)
        makeNetwork = getWhatToDo(args, RC)
        if makeNetwork:
            global G
            G = getNetwork(args, RC)
            G = getThresholds(args, G)
            outputNetwork(args, G)
    except KeyboardInterrupt:
        print('')
        return 0
    except Exception as e:
        if args.debug:
            raise e
        else:
            print("A {0} error occured: {1}".format(type(e).__name__ ,e))
            return e
    else:
        return 1

if __name__ == "__main__":
    mkCLI()



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
from .tagFunctions import tagToFunc
from .helpFuncs import reverseDict, makeBiDirectional

tagToFullDict = {k : v.__name__ for k, v in tagToFunc.items()}

fullToTagDict = reverseDict(tagToFullDict) #Reverses tagToFull

fullToTagDictUpper = {k.upper() : v for k,v in fullToTagDict.items()}


tagNameConverterDict = makeBiDirectional(tagToFullDict) #tagToFull made reversible

tagsAndNameSet = set(tagNameConverterDict.keys()) #set of WOS tags and their names

tagsAndNameSetUpper = set(([c.upper() for c in tagsAndNameSet]))

knownTagsList = list(tagToFullDict.keys()) #list of all the known tags

def tagToFull(tag):
    """A wrapper for `tagToFullDict`, it maps 2 character tags to their full names.

    # Parameters

    _tag_: `str`

    > A two character string giving the tag

    # Returns

    `str`

    > The full name of _tag_
    """
    try:
        return tagToFullDict[tag]
    except KeyError:
        raise("Tag not in list of known tags")


def normalizeToTag(val):
    """Converts tags or full names to 2 character tags, case insensitive

    # Parameters

    _val_: `str`

    > A two character string giving the tag or its full name

    # Returns

    `str`

    > The short name of _val_
    """
    try:
        val = val.upper()
    except AttributeError:
        raise KeyError("{} is not a tag or name string".format(val))
    if val not in tagsAndNameSetUpper:
        raise KeyError("{} is not a tag or name string".format(val))
    else:
        try:
            return fullToTagDictUpper[val]
        except KeyError:
            return val

def normalizeToName(val):
    """Converts tags or full names to full names, case sensitive

    # Parameters

    _val_: `str`

    > A two character string giving the tag or its full name

    # Returns

    `str`

    > The full name of _val_
    """
    if val not in tagsAndNameSet:
        raise KeyError("{} is not a tag or name string".format(val))
    else:
        try:
            return tagToFullDict[val]
        except KeyError:
            return val

def isTagOrName(val):
    """Checks if _val_ is a tag or full name of tag if so returns `True`

    # Parameters

    _val_: `str`

    > A string possible forming a tag or name

    # Returns

    `bool`

    > `True` if _val_ is a tag or name, otherwise `False`
    """
    return val in tagsAndNameSet



import collections
import io
import itertools

from ..mkExceptions import BadProQuestRecord, RecordsNotCompatible
from ..mkRecord import ExtendedRecord

from .tagProcessing.specialFunctions import proQuestSpecialTagToFunc
from .tagProcessing.tagFunctions import proQuestTagToFunc

class ProQuestRecord(ExtendedRecord):
    """Class for full ProQuest entries.

    This class is an [ExtendedRecord](./ExtendedRecord.html#metaknowledge.ExtendedRecord) capable of generating its own id number. You should not create them directly, but instead use [proQuestParser()](../modules/proquest.html#metaknowledge.proquest.proQuestHandlers.proQuestParser) on a ProQuest file.
    """
    def __init__(self, inRecord, recNum = None, sFile = "", sLine = 0):
        bad = False
        error = None
        fieldDict = None
        try:
            if isinstance(inRecord, dict) or isinstance(inRecord, collections.OrderedDict):
                fieldDict = collections.OrderedDict(inRecord)
            elif isinstance(inRecord, enumerate) or isinstance(inRecord, itertools.chain):
                #Already enumerated
                #itertools.chain is for the parser upstream to insert stuff into the stream
                fieldDict = proQuestRecordParser(inRecord, recNum)
            elif isinstance(inRecord, io.IOBase):
                fieldDict = proQuestRecordParser(enumerate(inRecord), recNum)
            elif isinstance(inRecord, str):
                #Probaly a better way to do this but it isn't going to be used much, so no need to improve it
                def addCharToEnd(lst):
                    for s in lst:
                        yield s + '\n'
                fieldDict = proQuestRecordParser(enumerate(addCharToEnd(inRecord.split('\n')), start = 1), recNum)
                #string io
            else:
                raise TypeError("Unsupported input type '{}', ProQuestRecords cannot be created from '{}'".format(inRecord, type(inRecord)))
        except BadProQuestRecord as b:
            self.bad = True
            self.error = b
            fieldDict = collections.OrderedDict()
        try:
            self._proID = "PROQUEST:{}".format(fieldDict["ProQuest document ID"][0])
        except KeyError:
            self._proID = "PROQUEST:MISSING"
            bad = True
            error = BadProQuestRecord("Missing ProQuest document ID")
        ExtendedRecord.__init__(self, fieldDict, self._proID, bad, error, sFile =sFile, sLine = sLine)

    def encoding(self):
        return 'utf-8'

    @staticmethod
    def getAltName(tag):
        return None

    @staticmethod
    def tagProcessingFunc(tag):
        #Should not raise an exception
        #It might be faster to do this as a class attribute
        return proQuestTagToFunc(tag)

    def specialFuncs(self, key):
        return proQuestSpecialTagToFunc[key](self)
        #raise KeyError("There are no special functions given by default.")

    def writeRecord(self, infile):
        raise RecordsNotCompatible("ProQuest's data format cannot be written back to file. You can still write out a csv with writeCSV().")

def proQuestRecordParser(enRecordFile, recNum):
    """The parser [ProQuestRecords](../classes/ProQuestRecord.html#metaknowledge.proquest.ProQuestRecord) use. This takes an entry from [proQuestParser()](#metaknowledge.proquest.proQuestHandlers.proQuestParser) and parses it a part of the creation of a `ProQuestRecord`.

    # Parameters

    _enRecordFile_ : `enumerate object`

    > a file wrapped by `enumerate()`

    _recNum_ : `int`

    > The number given to the entry in the first section of the ProQuest file

    # Returns

    `collections.OrderedDict`

    > An ordered dictionary of the key-vaue pairs in the entry
    """
    tagDict = collections.OrderedDict()
    currentEntry = 'Name'
    while True:
        lineNum, line = next(enRecordFile)
        if line == '_' * 60 + '\n':
            break
        elif line == '\n':
            pass
        elif currentEntry is 'Name' or currentEntry is 'url':
            tagDict[currentEntry] = [line.rstrip()]
            currentEntry = None
        elif ':' in line and not line.startswith('http://'):
            splitLine = line.split(': ')
            currentEntry = splitLine[0]
            tagDict[currentEntry] = [': '.join(splitLine[1:]).rstrip()]
            if currentEntry == 'Author':
                currentEntry = 'url'
        else:
            tagDict[currentEntry].append(line.rstrip())
    return tagDict
